# -*- coding: utf-8 -*-

#  Copyright © 2013-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import deque, OrderedDict
from math import atan2, sin, cos, pi

import pybiklib.modelcommon
from pybiklib.modelcommon import epsdigits, epsilon


def roundeps(val):
    return round(val, epsdigits+1)
    
class Vector (pybiklib.modelcommon.Vector):
    def rounded(self):
        return self.__class__(roundeps(s) for s in self)
        
    def scaled(self, vector):
        return self.__class__(s*v for s,v in zip(self, vector))
        
    def equalfuzzy(self, other): return (self-other).length_squared() < epsilon
    def inversfuzzy(self, other): return (self+other).length_squared() < epsilon
        
        
def distance_axis_vert(axis, vert):
    ''' axis:  vector that defines a line through (0,0,0)
        return: the shortest distance between axis and vert
    '''
    axis = Vector(axis)
    base_point = axis * vert.point.dot(axis) / axis.dot(axis)
    return (vert.point - base_point).length()
    
def distance_axis_edge(axis, verts):
    ''' axis:  vector that defines a line through (0,0,0)
        verts: two points that define a line segment
        return: the shortest distance between axis and edge
    '''
    vec_axis = axis             # S1.P1 - S1.P0, S1.P0 == (0,0,0)
    vec_edge = verts[1].point - verts[0].point  # S2.P1 - S2.P0
    vec_vert0 = -verts[0].point  # S1.P0 - S2.P0
    a = vec_axis.dot(vec_axis)  # always >= 0
    b = vec_axis.dot(vec_edge)
    c = vec_edge.dot(vec_edge)  # always >= 0
    d = vec_axis.dot(vec_vert0)
    e = vec_edge.dot(vec_vert0)
    D = a*c - b*b   # always >= 0
    sD = D      # sc = sN / sD, default sD = D >= 0
    tD = D      # tc = tN / tD, default tD = D >= 0
    
    # compute the line parameters of the two closest points
    if D < epsilon: # the lines are almost parallel
        sN = 0.0    # force using point P0 on segment S1
        sD = 1.0    # to prevent possible division by 0.0 later
        tN = e
        tD = c
    else:   # get the closest points on the infinite lines
        sN = b*e - c*d
        tN = a*e - b*d
        
    if tN < 0.0:  # tc < 0 => the t=0 edge is visible
        tN = 0.0
        # recompute sc for this edge
        sN = -d
        sD = a
    elif tN > tD:   # tc > 1  => the t=1 edge is visible
        tN = tD
        # recompute sc for this edge
        sN = (-d + b)
        sD = a
    # finally do the division to get sc and tc
    sc = 0.0 if abs(sN) < epsilon else sN / sD
    tc = 0.0 if abs(tN) < epsilon else tN / tD
    
    # get the difference of the two closest points
    dP = vec_vert0 + (vec_axis * sc) - (vec_edge * tc)    # =  S1(sc) - S2(tc)
    
    return dP.length()     # return the closest distance
    
    
class Plane:
    def __init__(self, normal):
        self.normal = Vector(normal)
        self.distance = 0
        
    def __repr__(self):
        return '{0}({1.normal}, {1.distance})'.format(self.__class__.__name__, self)
        
    def signed_distance(self, point):
        return self.normal.dot(point) - self.distance
        
    def intersect_segment(self, p, q):
        pvalue = self.signed_distance(p.point)
        qvalue = self.signed_distance(q.point)
        if abs(pvalue - qvalue) < epsilon:
            return -1, None
        if abs(pvalue) < epsilon:
            return 0, p
        if abs(qvalue) < epsilon:
            return 0, q
        if pvalue < 0 < qvalue or qvalue < 0 < pvalue:
            d = pvalue / (pvalue - qvalue)
            return 1, Vert(p.point + (q.point - p.point) * d)
        else:
            return -1, None
            
        
class GeomBase:
    def __repr__(self):
        def args():
            for attr in self.fields_arg:
                yield str(list(getattr(self, attr)))
            for attr in self.fields_kwarg:
                yield '{}={!r}'.format(attr, getattr(self, attr))
        return '{}({})'.format(self.__class__.__name__, ', '.join(args()))
        
    def __euler(self):
        objnames = ['verts', 'edges', 'faces', 'cells']
        strx = ''
        euler = 0
        for i, objname in zip(range(self.dim), objnames):
            count = sum(1 for obj in getattr(self, objname))
            strx += ' {}={}'.format(objname[0], count)
            euler += -count if i % 2 else count
        euler += -1 if self.dim % 2 else 1
        return 'x={}{} {}=1'.format(euler, strx, objnames[self.dim][0])
        
    def __genstr(self, indent=''):
        def args():
            for attr in self.fields_kwarg:
                yield '{}={!r}'.format(attr, getattr(self, attr))
        yield '{}{}: {} {}'.format(indent, self.__class__.__name__, ' '.join(args()), self.__euler())
        for attr in self.fields_arg:
            for v in getattr(self, attr):
                yield from v.__genstr(indent+'  ')
                    
    def __str__(self):
        return '\n'.join(self.__genstr())
        
    def clone(self, clone_data):
        if id(self) in clone_data:
            return
        obj = clone_data[id(self)] = self.__class__()
        vals = [(attr, getattr(self, attr)) for attr in reversed(self.fields_geom)]
        def func(clone_data):
            for attr in self.fields_kwarg:
                if attr not in self.fields_geom:
                    setattr(obj, attr, getattr(self, attr))
            for attr, val in vals:
                newval = [clone_data[id(o)] for o in val] if isinstance(val, list) else clone_data[id(val)]
                setattr(obj, attr, newval)
        for attr, val in vals:
            if isinstance(val, list):
                for o in val:
                    yield o.clone
            else:
                yield val.clone
        yield func
        
    def center(self):
        c = Vector()
        for i, v in enumerate(self.verts):
            c += v.point
        return c / (i+1)
        
        
class Vert (GeomBase):
    dim = 0
    fields_arg = ()
    fields_kwarg = 'point', 'id',
    fields_geom = ()
    
    def __init__(self, point=None, *, id=None):
        self.point = Vector() if point is None else Vector(point)
        self.id = id
        
    def assert_valid(self):
        import numbers
        assert len(self.point) == 3
        assert all(isinstance(i, numbers.Number) for i in self.point)
        return True
        
        
class HalfEdge (GeomBase):
    dim = 1
    fields_arg = 'verts',
    fields_kwarg = 'edge',
    fields_geom = 'verts', 'edge', '_nextatface', 'other', 'halfface'
    
    def __init__(self, verts=None, edge=None):
        if verts is None and edge is None:
            return
        self.verts = list(verts)
        self.edge = edge
        if edge is not None:
            edge.halfedges.append(self)
        self._nextatface = None
        self.other = None # other halfedge of edge in the same cell
        self.halfface = None
        
    @classmethod
    def from_other(cls, halfedge):
        he = cls(reversed(halfedge.verts), halfedge.edge)
        halfedge.other = he
        he.other = halfedge
        return he
        
    def assert_valid(self):
        assert all(isinstance(v, Vert) for v in self.verts)
        assert isinstance(self.edge, Edge)
        assert isinstance(self._nextatface, self.__class__)
        assert isinstance(self.other, self.__class__), self.other.__class__
        assert isinstance(self.halfface, HalfFace)
        assert all(v.assert_valid() for v in self.verts)
        assert len(self.verts) == 2
        assert self.edge.assert_valid()
        assert self._nextatface is not None
        assert self._nextatface is not self # len > 1
        assert self._nextatface._nextatface is not self # len > 2
        assert self._nextatface.edge is not self.edge
        assert self.verts[1] is self._nextatface.verts[0]
        assert self.other.other == self
        assert self.halfface is self._nextatface.halfface
        return True
        
    def __iter__(self):
        for v in self.verts:
            yield v
            
    @property
    def prevatface(self):
        he = self
        while he._nextatface != self:
            he = he._nextatface
        return he
        
    def addtoface(self, halfedge, next=None):
        halfedge._nextatface = next or self._nextatface
        halfedge.halfface = self.halfface
        self._nextatface = halfedge
        
    def bevel_point(self, dist):
        '''
            >>> poly = Polyhedron()
            >>> poly.set_verts([0.,0.,0.], [1.,0.,0.], [1., 2.,0.])
            >>> poly.set_edges([0,1], [1,2], [2,0])
            >>> poly.set_faces([1,2,3])
            >>> poly.edges[0].halfedges[0].bevel_point(.1)
            Vert(0.9, 0.1, 0.0)
        '''
        vert1, vert2 = self.verts
        vert3 = self._nextatface.verts[1]
        nd1 = (vert1.point - vert2.point).normalised()
        nd2 = (vert3.point - vert2.point).normalised()
        return vert2.point + (nd1 + nd2) * dist / nd1.cross(nd2).length()
        
        
class Edge (GeomBase):
    fields_arg = ()
    fields_kwarg = 'id',
    fields_geom = 'halfedges',
    
    def __init__(self, *, id=None):
        self.halfedges = []
        self.id = id
        
    def assert_valid(self):
        assert all(isinstance(he, HalfEdge) for he in self.halfedges)
        assert len(self.halfedges) > 0
        assert all(isinstance(he, HalfEdge) for he in self.halfedges)
        assert all(he.edge is self for he in self.halfedges)
        assert all(he._nextatface is not None for he in self.halfedges)
        return True
        
    def __str__(self):
        verts = len(self.halfedges)
        euler = verts - 1
        return 'χ={} v={} e=1'.format(euler, verts)
        
    def split(self, vert):
        verts = self.halfedges[0].verts
        edge2 = self.__class__()
        for he1 in self.halfedges:
            if he1.verts[0] == verts[0]:
                he2 = HalfEdge([vert, he1.verts[1]], edge2)
                he1.verts[1] = vert
                he1.addtoface(he2)
                he2_other = he1.other.prevatface
            else:
                assert he1.verts[1] == verts[0], (he1.verts, verts)
                he2 = HalfEdge([he1.verts[0], vert], edge2)
                he1.verts[0] = vert
                he1.prevatface.addtoface(he2)
                he2_other = he1.other._nextatface
            if he2_other.edge is edge2:
                he2.other = he2_other
                he2_other.other = he2
                
        
class HalfFace (GeomBase):
    dim = 2
    fields_arg = 'halfedges',
    fields_kwarg = 'face',
    fields_geom = 'halfedge', '_nextatcell', 'face'
    
    def __init__(self, *, halfedges=None, face=None):
        if face is not None:
            self.face = face
            self.face.halffaces.append(self)
        if halfedges:
            self.set_halfedges(halfedges)
        else:
            self.halfedge = None
        self._nextatcell = None
        
    def assert_valid(self):
        assert isinstance(self.face, Face)
        assert self.face.assert_valid()
        assert isinstance(self._nextatcell, self.__class__)
        assert self._nextatcell is not None
        assert self.halfedge.halfface is self
        assert all(isinstance(he, HalfEdge) for he in self.halfedges)
        assert all(he.assert_valid() for he in self.halfedges)
        return True
        
    def __iter__(self):
        if self.halfedge is None:
            return
        he = self.halfedge
        while True:
            yield he
            he = he._nextatface
            if he is self.halfedge:
                return
                
    halfedges = property(__iter__)
        
    def halfedges2(self):
        if self.halfedge is None:
            return
        he = self.halfedge
        while True:
            yield he, he._nextatface
            he = he._nextatface
            if he is self.halfedge:
                return
                
    @property
    def edges(self):
        for he in self.halfedges:
            yield he.edge
            
    @property
    def verts(self):
        for he in self.halfedges:
            yield he.verts[0]
            
    def set_halfedges(self, halfedges):
        ithalfedges = iter(halfedges)
        self.halfedge = phe = next(ithalfedges)
        for he in ithalfedges:
            phe._nextatface = he
            phe.halfface = self
            phe = he
        phe._nextatface = self.halfedge
        phe.halfface = self
        
    def normal(self):
        verts = self.verts
        try:
            v1 = next(verts).point
            v2 = next(verts).point
            v3 = next(verts).point
        except StopIteration:
            raise ValueError('Too few vertices')
        return (v2-v1).cross(v3-v1).normalised()
        
    @staticmethod
    def _sorted_halfedges(halfedges):
        try:
            current = halfedges.pop(0)
        except IndexError:
            return halfedges
        halfedges_new = [current]
        unused, prev_vert = current.verts
        while halfedges:
            for i, he in enumerate(halfedges):
                if prev_vert not in he.verts:
                    continue
                current = halfedges.pop(i)
                assert current is he
                vert1, vert2 = he.verts
                if prev_vert is vert2:
                    current.verts = [vert2, vert1]
                    prev_vert = vert1
                else:
                    prev_vert = vert2
                halfedges_new.append(current)
                break
            else:
                assert False
        halfedges[:] = halfedges_new
        return halfedges
        
    def reverse(self):
        halfedges = list(self.halfedges)
        for he in halfedges:
            he.verts = list(reversed(he.verts))
        self._sorted_halfedges(halfedges)
        self.set_halfedges(halfedges)
        
        
class Face (GeomBase):
    fields_arg = ()
    fields_kwarg = 'type', 'id'
    fields_geom = 'halffaces',
    
    def __init__(self, *, type=None, id=None):
        self.halffaces = []
        self.type = type
        self.id = id
        
    def assert_valid(self):
        assert all(isinstance(f, HalfFace) for f in self.halffaces)
        assert all(hf.face is self for hf in self.halffaces)
        assert 1 <= len(self.halffaces) <= 2
        lenhf0 = len(list(self.halffaces[0].halfedges))
        assert all(lenhf0 == len(list(hf.halfedges)) for hf in self.halffaces)
        return True
        
    def __str__(self):
        verts = len(list(self.verts()))
        edges = len(list(self.edges()))
        euler = verts - edges + 1
        return 'χ={} v={} e={} f=1'.format(euler, verts, edges)
        
    @property
    def edges(self):
        return self.halffaces[0].edges
        
    @property
    def verts(self):
        return self.halffaces[0].verts
        
    def remove_all_edges(self):
        for hf in self.halffaces:
            hf.halfedge = None
            
    def split(self, split_verts):
        assert len(split_verts) == 2, len(split_verts)
        edge = Edge()
        newface = Face(type=self.type, id=self.id)
        edgef1 = None
        for halfface in self.halffaces:
            he = halfface.halfedge
            # make sure to start on the same edge for all halffaces
            if edgef1 is None:
                edgef1 = he.edge
            else:
                while he.edge is not edgef1:
                    he = he._nextatface
                halfface.halfedge = he
            # proceed to the end of the old halfface
            while he.verts[1] not in split_verts:
                he = he._nextatface
            hedge1f1 = he
            he = he._nextatface
            hedge1f2 = he
            # create a new halfface
            newhalfface = HalfFace(face=newface)
            newhalfface.halfedge = he
            newhalfface._nextatcell = halfface._nextatcell
            halfface._nextatcell = newhalfface
            # proceed to the end of the new halfface
            while he.verts[1] not in split_verts:
                he.halfface = newhalfface
                he = he._nextatface
            he.halfface = newhalfface
            hedge2f2 = he
            he = he._nextatface
            hedge2f1 = he
            # insert a new halfedge between the splitted halfface
            hedge1 = HalfEdge(split_verts, edge)
            hedge2 = HalfEdge([split_verts[1], split_verts[0]], edge)
            hedge1.other = hedge2
            hedge2.other = hedge1
            if hedge1f1.verts[1] is split_verts[1]:
                hedge1, hedge2 = hedge2, hedge1
            hedge1f1.addtoface(hedge1, hedge2f1)
            hedge2f2.addtoface(hedge2, hedge1f2)
        return edge
        
    @classmethod
    def polygon(cls, point, n, ids=None):
        x0, y0 = point
        vert0 = Vert(point)
        r = vert0.point.length()
        angle_v0 = atan2(y0, x0)
        angle_diff = 2 * pi / n
        
        vertp = vert0
        halfedges = []
        for k in range(1, n):
            angle = k * angle_diff + angle_v0
            vertc = Vert((r * cos(angle), r * sin(angle)))
            edge = Edge() if ids is None else Edge(id=ids[k-1])
            halfedges.append(HalfEdge((vertp, vertc), edge))
            vertp = vertc
        edge = Edge() if ids is None else Edge(id=ids[n-1])
        halfedges.append(HalfEdge((vertp, vert0), edge))
        return HalfFace(halfedges=halfedges, face=Face()).face
        
    def extrude_y(self, upy, downy, ids=None):
        assert len(self.halffaces) == 1
        halfface_up = self.halffaces[0]
        halfedges_down = []
        halffaces = [halfface_up]
        
        def extrude_vert(vert_up, upy, downy):
            x1, y1 = vert_up.point
            vert_up.point = Vector((x1, upy, -y1))
            vert_down = Vert((x1, downy, -y1))
            return vert_up, vert_down
            
        vp_down = None
        for he_up in halfface_up.halfedges:
            # vertices
            vc_up, vc_down = extrude_vert(he_up.verts[1], upy, downy)
            # edges
            e_down = Edge()
            ec_side = Edge()
            if vp_down is None:
                v1_up, v1_down = vc_up, vc_down
                e0_down = e_down
                e1_side = ec_side
            else:
                # down halfedge
                halfedges_down.append(HalfEdge([vc_down, vp_down], e_down))
                # side face
                he1 = HalfEdge([vp_down, vc_down], e_down)
                he2 = HalfEdge([vc_down, vc_up], ec_side)
                he3 = HalfEdge([vc_up, vp_up], he_up.edge)
                he4 = HalfEdge([vp_up, vp_down], ep_side)
                halfface_side = HalfFace(halfedges=[he1, he2, he3, he4], face=Face(type='face', id=he_up.edge.id))
                halffaces.append(halfface_side)
            # next
            vp_up, vp_down = vc_up, vc_down
            ep_side = ec_side
        he_up = halfface_up.halfedge
        # joining down halfedge
        halfedges_down.append(HalfEdge([v1_down, vp_down], e0_down))
        # joining side face
        he1 = HalfEdge([vp_down, v1_down], e0_down)
        he2 = HalfEdge([v1_down, v1_up], e1_side)
        he3 = HalfEdge([v1_up, vp_up], he_up.edge)
        he4 = HalfEdge([vp_up, vp_down], ep_side)
        halfface_side = HalfFace(halfedges=[he1, he2, he3, he4], face=Face(type='face', id=he_up.edge.id))
        halffaces.append(halfface_side)
        # up and down faces
        face_down = Face()
        if ids is not None:
            self.type = face_down.type = 'face'
            self.id, face_down.id = ids
        halfface_down = HalfFace(halfedges=reversed(halfedges_down), face=face_down)
        halffaces.append(halfface_down)
        return Cell(halffaces=halffaces)
        
    def pyramid(self, upy, downy, id=None):
        self.assert_valid()
        assert len(self.halffaces) == 1
        halfface_down = self.halffaces[0]
        halffaces = [halfface_down]
        vert_up = Vert((0., upy, 0.))
        
        def change_vert(vert_down, downy):
            x1, y1 = vert_down.point
            vert_down.point = Vector((-x1, downy, -y1))
            return vert_down
            
        vp_down = None
        for he in halfface_down.halfedges:
            # vertices
            vc_down = change_vert(he.verts[1], downy)
            # edges
            ec_side = Edge()
            if vp_down is None:
                v1_down = vc_down
                e1_side = ec_side
            else:
                # side face
                he1 = HalfEdge([vc_down, vp_down], he.edge)
                he2 = HalfEdge([vp_down, vert_up], ep_side)
                he3 = HalfEdge([vert_up, vc_down], ec_side)
                halfface_side = HalfFace(halfedges=[he1, he2, he3], face=Face(type='face', id=he.edge.id))
                halffaces.append(halfface_side)
            # next
            vp_down = vc_down
            ep_side = ec_side
        he = halfface_down.halfedge
        # joining side face
        he1 = HalfEdge([v1_down, vp_down], he.edge)
        he2 = HalfEdge([vp_down, vert_up], ep_side)
        he3 = HalfEdge([vert_up, v1_down], e1_side)
        halfface_side = HalfFace(halfedges=[he1, he2, he3], face=Face(type='face', id=he.edge.id))
        halffaces.append(halfface_side)
        # down face
        if id is not None:
            self.type = 'face'
            self.id = id
        return Cell(halffaces=halffaces)
        
        
class Cell (GeomBase):
    dim = 3
    fields_arg = 'halffaces',
    fields_kwarg = 'indices',
    fields_geom = 'halfface',
    
    def __init__(self, *, halffaces=None):
        if halffaces is not None:
            phf = halffaces[-1]
            for hf in halffaces:
                phf._nextatcell = hf
                phf = hf
                for he in hf.halfedges:
                    if len(he.edge.halfedges) == 2: #XXX:
                        he.edge.halfedges[0].other = he.edge.halfedges[1]
                        he.edge.halfedges[1].other = he.edge.halfedges[0]
            self.halfface = halffaces[0]
        self.indices = None
        
    def assert_valid(self):
        assert all(isinstance(f, HalfFace) for f in self.halffaces)
        assert all(hf.assert_valid() for hf in self.halffaces)
        assert self.halfface is not None # len > 0
        assert self.halfface is not self # len > 1
        assert self.halfface._nextatcell is not self # len > 2
        assert self.halfface._nextatcell._nextatcell is not self # len > 3
        assert len(list(self.edges)) > 3
        assert len(list(self.verts)) > 3
        assert all(hf.halfedge.other.halfface in self.halffaces for hf in self.halffaces)
        return True
        
    def __iter__(self):
        if self.halfface is None:
            return
        hf = self.halfface
        while True:
            yield hf
            hf = hf._nextatcell
            if hf is self.halfface:
                return
                
    halffaces = property(__iter__)
        
    @property
    def faces(self):
        for hf in self.halffaces:
            yield hf.face
            
    @property
    def edges(self):
        cache = []
        for f in self.faces:
            for e in f.edges:
                if e not in cache:
                    cache.append(e)
                    yield e
                    
    @property
    def verts(self):
        cache = []
        for hf in self.halffaces:
            for v in hf.verts:
                if v not in cache:
                    cache.append(v)
                    yield v
                    
    def add(self, halfface):
        halfface._nextatcell = self.halfface._nextatcell
        self.halfface._nextatcell = halfface
        
    def split(self, split_edges, id=None):
        halffaces, halffaces1, halffaces2 = [self.halfface], [self.halfface], []
        halfedges1, halfedges2 = [], []
        while halffaces:
            for he in halffaces.pop().halfedges:
                hf = he.other.halfface
                if he.edge in split_edges:
                    halfedges1.append(HalfEdge.from_other(he))
                    if hf not in halffaces2:
                        halffaces2.append(hf)
                else:
                    if hf not in halffaces1:
                        halffaces.append(hf)
                        halffaces1.append(hf)
        assert len(halfedges1) > 2
        halffaces = halffaces2[:]
        while halffaces:
            for he in halffaces.pop().halfedges:
                if he.edge in split_edges:
                    halfedges2.append(HalfEdge.from_other(he))
                else:
                    hf = he.other.halfface
                    if hf not in halffaces2:
                        halffaces.append(hf)
                        halffaces2.append(hf)
        assert len(halfedges2) > 2
        # new face
        face = Face(type='cut', id=id)
        halfface1 = HalfFace(halfedges=HalfFace._sorted_halfedges(halfedges1), face=face)
        halfface2 = HalfFace(halfedges=HalfFace._sorted_halfedges(halfedges2), face=face)
        # update cell
        phf = halfface1
        for hf in halffaces1:
            phf._nextatcell = hf
            phf = hf
        phf._nextatcell = halfface1
        # new cell
        halffaces2.append(halfface2)
        new_cell = Cell(halffaces=halffaces2)
        return new_cell
            
        
class Polyhedron:
    def __init__(self):
        self._cells = []
        
    def clone(self):
        clone_data = {}
        obj = self.__class__()
        funcs = deque(f for o in self._cells for f in o.clone(clone_data))
        obj._cells = [clone_data[id(o)] for o in self._cells]
        while funcs:
            f = funcs.popleft()
            r = f(clone_data)
            if r is not None:
                funcs.extendleft(reversed(list(r)))
        return obj
        
    def assert_valid(self):
        import itertools
        assert all(isinstance(c, Cell) and c.assert_valid() for c in self.cells)
        assert all(isinstance(f, Face) and f.assert_valid() for f in self.faces)
        assert all(isinstance(e, Edge) and e.assert_valid() for e in self.edges)
        assert all(isinstance(v, Vert) and v.assert_valid() for v in self.verts)
        assert all(not v1.equalfuzzy(v2) for v1, v2 in itertools.combinations(self.verts, 2)), self.verts
        return True
        
    def __str__(self):
        verts = len(self.verts)
        edges = len(self.edges)
        faces = len(self.faces)
        cells = len(self.cells)
        euler = verts - edges + faces - cells
        return 'χ={} v={} e={} f={} c={}'.format(euler, verts, edges, faces, cells)
        
    def verts_string(self, verts): return '-'.join(str(self.verts.index(v)) for v in verts)
    
    @property
    def cells(self):
        return self._cells
    @cells.setter
    def cells(self, cells):
        self._cells = cells
        
    @property
    def faces(self):
        cache = []
        for c in self.cells:
            for f in c.faces:
                if f not in cache:
                    cache.append(f)
        return cache
        
    @property
    def edges(self):
        cache = []
        for f in self.faces:
            for e in f.edges:
                if e not in cache:
                    cache.append(e)
        return cache
        
    @property
    def verts(self):
        cache = []
        for f in self.faces:
            for v in f.verts:
                if v not in cache:
                    cache.append(v)
        return cache
        
    def set_verts(self, *args):
        self._polytopes = [Vert(arg) for arg in args]
            
    def set_edges(self, *args):
        halfedges = []
        for arg in args:
            verts = [self._polytopes[i] for i in arg]
            halfedges.append(HalfEdge(verts=verts, edge=Edge()))
        self._polytopes = halfedges
        
    def set_faces(self, *args, ids=None):
        halffaces = []
        if ids is None:
            faces = [Face() for unused in args]
        else:
            assert len(args) == len(ids)
            faces = [Face(type='face', id=fid) for fid in ids]
        for arg, face in zip(args, faces):
            halfedges = []
            for i in arg:
                # this only works if args describe one closed cell
                if i > 0:
                    he = self._polytopes[i-1]
                elif i < 0:
                    he = HalfEdge.from_other(self._polytopes[-i-1])
                else:
                    raise ValueError()
                halfedges.append(he)
            halfface = HalfFace(halfedges=halfedges, face=face)
            halffaces.append(halfface)
        del self._polytopes
        self._cells = [Cell(halffaces=halffaces)]
            
    _next_split_id = 0
        
    def split_plane(self, plane, indexpos=None):
        # split intersected edges
        new_verts = []
        for edge in self.edges[:]:
            pos, vert = plane.intersect_segment(*edge.halfedges[0].verts)
            if pos > 0:
                edge.split(vert)
                new_verts.append(vert)
            elif pos == 0 and vert not in new_verts:
                new_verts.append(vert)
        # split intersected faces
        new_edges = []
        for face in self.faces[:]:
            split_verts = [v for v in face.verts for nv in new_verts if nv is v]
            assert 0 <= len(split_verts) <= 2, split_verts
            if len(split_verts) == 2:
                for he in face.halffaces[0].halfedges:
                    if he.verts == split_verts or he.verts == [split_verts[1], split_verts[0]]:
                        new_edges.append(he.edge)
                        break
                else:
                    new_edge = face.split(split_verts)
                    new_edges.append(new_edge)
        # split cells
        split_id = self._next_split_id
        self.__class__._next_split_id += 1
        for cell in self.cells[:]:
            split_edges = [e for e in cell.edges if e in new_edges]
            if len(split_edges) > 2:
                new_cell = cell.split(split_edges, split_id)
                self.cells.append(new_cell)
            else:
                new_cell = None
            if indexpos is not None:
                for hf in cell.halffaces:
                    for he in hf.halfedges:
                        if he.edge not in split_edges:
                            sd1 = -plane.signed_distance(he.verts[0].point)
                            # the verts of the edge e cannot be both in range epsilon to the plane,
                            # otherwise e would be in split_edges, so only test one vert for epsilon
                            if sd1 > epsilon or sd1 >= -epsilon and plane.signed_distance(he.verts[1].point) < 0:
                                if new_cell is not None:
                                    new_cell.indices = cell.indices
                                new_cell = cell
                            elif new_cell is None:
                                break
                            indices = list(cell.indices)
                            indices[indexpos] += 1
                            new_cell.indices = tuple(indices)
                            break
                    else:
                        continue
                    break
        return split_id
            
    def scale(self, value):
        for v in self.verts:
            v.point = v.point.scaled(value)
            
    def bevel(self, width):
        faces = self.faces
        
        f_vverts_eedges = {}
        for f in faces:
            vverts = {id(he.verts[1]): Vert(he.bevel_point(width), id=f.id) for he in f.halffaces[0].halfedges}
            eedges = {id(e): Edge() for e in f.edges}
            f_vverts_eedges[f] = (vverts, eedges)
        for c in self.cells:
            # create the faces that replace the old edges
            vhalfedges = OrderedDict()  # for reproducible build
            for he1 in (he for hf in c.halffaces for he in hf.halfedges):
                if he1.halfface.face.type == 'bevel':
                    continue
                he3 = he1.other
                if he3.halfface.face.type == 'bevel':
                    continue
                vverts1, eedges1 = f_vverts_eedges[he1.halfface.face]
                vverts3, eedges3 = f_vverts_eedges[he3.halfface.face]
                idhe1v2 = id(he1.verts[1])
                idhe3v2 = id(he3.verts[1])
                # new halfedges for the adjacent vertex-faces
                he2 = HalfEdge([], edge=Edge())
                he4 = HalfEdge([], edge=Edge())
                vhalfedges.setdefault(idhe1v2, []).append(he2)
                vhalfedges.setdefault(idhe3v2, []).append(he4)
                # update halfedges for the adjacent faces
                ide = id(he1.edge)
                edge = eedges1[ide]
                he1.edge = edge
                edge.halfedges.append(he1)
                edge = eedges3[ide]
                he3.edge = edge
                edge.halfedges.append(he3)
                # replace the adjacent verts
                vert1 = vverts1[idhe3v2]
                vert2 = vverts1[idhe1v2]
                vert3 = vverts3[idhe1v2]
                vert4 = vverts3[idhe3v2]
                he1.verts[:] = [vert1, vert2]
                he2.verts[:] = [vert2, vert3]
                he3.verts[:] = [vert3, vert4]
                he4.verts[:] = [vert4, vert1]
                # new halfedges for the new face
                he1 = HalfEdge.from_other(he1)
                he2 = HalfEdge.from_other(he2)
                he3 = HalfEdge.from_other(he3)
                he4 = HalfEdge.from_other(he4)
                c.add(HalfFace(halfedges=[he4, he3, he2, he1], face=Face(type='bevel')))
            # create the faces that replace the old verts
            for halfedges in vhalfedges.values():
                c.add(HalfFace(halfedges=HalfFace._sorted_halfedges(halfedges), face=Face(type='bevel')))
                    
    def remove_faces(self, func):
        for cell in self.cells:
            halffaces = list(cell.halffaces)
            phf = halffaces[-1]
            for hf in halffaces:
                if func(hf.face):
                    phf._nextatcell = hf._nextatcell
                    hf._nextatcell = None
                    hf.face.halffaces.remove(hf)
                    hf.face = None
                    if hf is cell.halfface:
                        cell.halfface = phf._nextatcell
                else:
                    phf = hf
        
    def remove_cells(self, func):
        self._cells = [c for c in self._cells if not func(c)]
        
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2009, 2011-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# pylint: disable=W0201


from glob import glob
import os
import sys

use_setuptools = not os.getenv('PYBIK_NO_SETUPTOOLS')
if use_setuptools:
    try:
        import setuptools
    except ImportError:
        use_setuptools = False
if not use_setuptools:
    from distutils.core import setup
    from distutils.cmd import Command
    from distutils.extension import Extension
    from distutils import dist
    from distutils.command.install import install as install_orig
else:
    print('using setuptools')
    from setuptools import setup
    from setuptools import Command
    from setuptools import Extension
    from setuptools import dist
    from setuptools.command.install import install as install_orig

from distutils.dir_util import remove_tree
from distutils.file_util import move_file
from distutils.dep_util import newer, newer_group
from distutils.log import warn, info, error, debug

import distutils.command.build
import distutils.command.build_ext
import distutils.command.build_scripts
import distutils.command.build_py
import distutils.command.install_data
import distutils.command.install_lib
import distutils.command.clean
import distutils.command.sdist

#from Cython.Distutils import build_ext

#print('using', 'setuptools' if use_setuptools else 'distutils')

# the clean command should leave a clean tree
sys.dont_write_bytecode = True
os.environ['PYTHONDONTWRITEBYTECODE'] = '1'

from pybiklib import config


class Distribution (distutils.dist.Distribution):
    def __init__(self, attrs=None):
        self.bug_contact = None
        self.qt_ui_files = None
        self.desktop_templates = None
        #XXX: Mangle alpha/beta/rc versions to PEP440 version scheme for setuptools, pip, pypi ...
        version = attrs['version']
        pep440 = version.replace('~alpha', 'a').replace('~beta', 'b').replace('~rc', 'rc')
        attrs['version'] = pep440
        distutils.dist.Distribution.__init__(self, attrs=attrs)
        if version != pep440:
            self.metadata.version = version


class build(distutils.command.build.build):
    """Adds extra commands to the build target."""
    user_options = distutils.command.build.build.user_options + [
            ('inplace', 'i', 'ignore build-lib and put compiled modules into the source directory'),
            #XXX: In Python 3.5 distutils.command.build.build has a parallel option
            ('parallel=', None, 'number of parallel build jobs or "True"'),
            ('arch-only', None, 'Build only architecture dependent files'),
            ('indep-only', None, 'Build only architecture independent files'),
            ('fast', None, 'Build models without optimization'),
            ('pickle-protocol=', None, 'Pickle protocol for modeldata (default: highest protocol)'),
            ('maxsize=', None, 'Maximum model size (default: 10)'),
            ('reproducible', None, 'Build modeldata reproducible (experimental)'),
        ]
    boolean_options = distutils.command.build.build.boolean_options + [
            'inplace', 'arch-only', 'indep-only', 'fast', 'reproducible',
        ]
        
    def has_pure_modules(self):
        return self.indep_only and distutils.command.build.build.has_pure_modules(self)
        
    _commands_dict = dict(distutils.command.build.build.sub_commands)
    # pylint: disable=W0212
    sub_commands = [('build_py',      has_pure_modules),
                    ('build_ext',     lambda self: self.arch_only and self._commands_dict['build_ext'](self)),
                    ('build_scripts', lambda self: self.indep_only and self._commands_dict['build_scripts'](self)),
                    ('build_models',  lambda self: self.indep_only),
                    ('build_i18n',    lambda self: self.indep_only),
                    ('build_desktop', lambda self: self.indep_only),
                    ('build_man',     lambda self: self.indep_only and not self.inplace),
                ]
    # pylint: enable=W0212
                
    def initialize_options(self):
        distutils.command.build.build.initialize_options(self)
        self.inplace = 0
        self.parallel = None
        self.arch_only = False
        self.indep_only = False
        self.fast = False
        self.pickle_protocol = '-1'
        self.maxsize = None
        self.reproducible = False
        
    def finalize_options(self):
        # In Python 3.5 distutils.command.build.build has a parallel option
        # convert 'true' to True before the base class fails
        if isinstance(self.parallel, str) and self.parallel.lower() == 'true':
            self.parallel = True
        distutils.command.build.build.finalize_options(self)
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
        
        
class build_csrc (distutils.command.build_ext.build_ext):
    description = 'build C source code with py2pyx and cython'
    user_options = distutils.command.build_ext.build_ext.user_options + [
        ('cython-opts=', None,          'Cython options'),
        ('gles=',        None,          'Build for GLES (experimental)'),
    ]
    
    def initialize_options(self):
        distutils.command.build_ext.build_ext.initialize_options(self)
        self.inplace = None
        self.cython_opts = ''
        self.gles = None
        
    def finalize_options(self):
        # deactivate parallel build in Python 3.5, does not work with this command
        self.parallel = 1
        distutils.command.build_ext.build_ext.finalize_options(self)
        self.set_undefined_options('build',
                                       ('inplace', 'inplace'),
                                    )
        self.cython_opts = self.cython_opts.split()
        
    def compare_file(self, filename1, filename2):
        if self.force or not os.path.exists(filename2):
            return False
        with open(filename1, 'rb') as file1, open(filename2, 'rb') as file2:
            return file1.read() == file2.read()
            
    @staticmethod
    def run_py2pyx(py_file, pyx_file, pxd_file):
        info('py2pyx: %s --> %s, %s' % (py_file, pyx_file, pxd_file))
        from buildlib.py2pyx import create_pyx, Py2pyxParseError
        try:
            create_pyx(py_file, pyx_file, pxd_file)
        except Py2pyxParseError as e:
            error("error: %s", e)
            sys.exit(1)
        
    identifier_translate = bytes((i if bytes([i]).isalnum() else ord('_')) for i in range(256))
    
    @classmethod
    def replace_cwd(cls, filename):
        with open(filename, 'rb') as f:
            text = f.read()
        cwdb_esc = bytes(ord(c) for c in os.getcwd()[:32] if ord(c)<128).translate(cls.identifier_translate)
        text = text.replace(os.getcwdb(), b'/tmp').replace(cwdb_esc, b'_tmp')
        with open(filename, 'wb') as f:
            f.write(text)
            
    def run_cython(self, cython_opts, infile, outfile):
        from Cython.Compiler.CmdLine import parse_command_line
        from Cython.Compiler.Main import compile
        
        info('cython: %s --> %s' % (infile, outfile))
        args = ['-3', '-o', outfile] + cython_opts + [infile]
        options, unused_sources = parse_command_line(args)
        try:
            result = compile([infile], options)
        except Exception as e:
            error("error: %s", e)
            sys.exit(1)
        if result.num_errors > 0:
            sys.exit(1)
        self.replace_cwd(outfile)
        
    def build_pyx(self, py_file, pyx_file, pxd_file):
        pxd_file_tmp = pxd_file + '.tmp'
        self.mkpath(os.path.dirname(pyx_file))
        self.make_file([py_file], pyx_file, self.run_py2pyx, (py_file, pyx_file, pxd_file_tmp))
        if os.path.exists(pxd_file_tmp):
            if self.compare_file(pxd_file_tmp, pxd_file):
                os.remove(pxd_file_tmp)
                info("unchanged file '%s'", pxd_file)
            else:
                if os.path.exists(pxd_file):
                    os.remove(pxd_file)
                move_file(pxd_file_tmp, pxd_file)
        
    def build_extension(self, extension):
        sources = extension.sources
        depends = sources + extension.depends
            
        # generate pxd- and pyx-files from py-files with #px annotations
        pyx_files = []
        pyx_files_dep = []
        for in_file in depends:
            base, ext = os.path.splitext(in_file)
            if ext == '.py':
                base = os.path.join(self.build_temp, os.path.splitext(in_file)[0])
                basename = os.path.basename(base)
                dirname = os.path.dirname(base)
                pyx_file = os.path.join(dirname, '_%s.pyx' % basename)
                pxd_file = os.path.join(dirname, '_%s.pxd' % basename)
                self.build_pyx(in_file, pyx_file, pxd_file)
                if in_file in extension.sources:
                    pyx_files.append(pyx_file)
                pyx_files_dep.append(pxd_file)
                
                if self.gles is not None:
                    info('Changing source for GLES %s', self.gles)
                    with open(pyx_file, 'rt', encoding='utf-8') as f:
                        text = f.read()
                    text = text.replace("DEF SOURCEGLVERSION = 'GL'", "DEF SOURCEGLVERSION = 'GLES%s'" % self.gles)
                    with open(pyx_file, 'wt', encoding='utf-8') as f:
                        f.write(text)
            elif ext in ('.pxi','.pxd'):
                out_file = os.path.join(self.build_temp, in_file)
                self.mkpath(os.path.dirname(out_file))
                self.copy_file(in_file, out_file)
                pyx_files_dep.append(out_file)
            else:
                if in_file in extension.sources:
                    pyx_files.append(in_file)
                else:
                    pyx_files_dep.append(in_file)
                    
        # generate C-files with cython
        if extension.language == 'c++':
            c_file_ext = '.cpp'
            cython_opts = ['--cplus'] + self.cython_opts
        else:
            c_file_ext = '.c'
            cython_opts = self.cython_opts
        c_files = []
        for in_file in pyx_files:
            base, ext = os.path.splitext(in_file)
            if ext == '.pyx':
                out_file = base + c_file_ext
                inplace_file = os.path.join('csrc', os.path.basename(out_file))
                if os.path.exists(inplace_file) and not self.force:
                    self.mkpath(os.path.dirname(out_file))
                    self.copy_file(inplace_file, out_file)
                else:
                    self.make_file(pyx_files+pyx_files_dep, out_file,
                                   self.run_cython, (cython_opts, in_file, out_file))
                c_files.append(out_file)
            else:
                c_files.append(in_file)
                
        # exclude Cython-files from dependencies
        c_files_dep = []
        for in_file in pyx_files_dep:
            base, ext = os.path.splitext(in_file)
            if ext not in ('.pxi', '.pxd'):
                c_files_dep.append(in_file)
        extension.depends = c_files_dep
        extension.sources = c_files
        
        
class build_ext (build_csrc):
    description = distutils.command.build_ext.build_ext.description
    
    def build_extension(self, extension):
        if self.gles == '2':
            extension.libraries[:] = ['GLESv2']
        sources = extension.sources
        ext_path = self.get_ext_fullpath(extension.name)
        depends = sources + extension.depends
        if self.force or newer_group(depends, ext_path, 'newer'):
            info('building C-Code for %r extension', extension.name)
        else:
            debug('skipping %r extension (up-to-date)', extension.name)
            return
            
        # build C-code
        build_csrc.build_extension(self, extension)
        # build extension module from C-code
        distutils.command.build_ext.build_ext.build_extension(self, extension)
        #HACK: Due to build_csrc.compare_file the C compiler may not run, even though
        #      the dependencies expect this. Therefore update the timestamp manually.
        ext_path = self.get_ext_fullpath(extension.name)
        try:
            os.utime(ext_path, None)
        except OSError:
            pass
        
        
class build_scripts (distutils.command.build_scripts.build_scripts):
    user_options = distutils.command.build_scripts.build_scripts.user_options + [
            ('inplace',     'i', 'ignore build-lib and put compiled modules into the source directory'),
            ('build-temp=', 't', "temporary build directory"),
        ]
    boolean_options = distutils.command.build.build.boolean_options + ['inplace']
    
    def initialize_options(self):
        distutils.command.build_scripts.build_scripts.initialize_options(self)
        self.inplace = None
        self.build_temp = None
        
    def finalize_options(self):
        distutils.command.build_scripts.build_scripts.finalize_options(self)
        self.set_undefined_options('build',
                    ('inplace', 'inplace'),
                    ('build_temp', 'build_temp'),
                )
                
    def run(self):
        build_dir = self.build_dir
        self.build_dir = self.build_temp
        distutils.command.build_scripts.build_scripts.run(self)
        self.build_dir = build_dir
        for script in self.scripts:
            outfile = os.path.basename(script)
            script = os.path.join(self.build_temp, outfile)
            if not os.path.exists(script):
                continue
            if not self.inplace:
                self.mkpath(self.build_dir)
                outfile = os.path.join(self.build_dir, outfile)
            outfile, ext = os.path.splitext(outfile)
            if ext != '.py':
                outfile += ext
            self.copy_file(script, outfile)
        
        
class build_py(distutils.command.build_py.build_py):
    description = 'build pure Python modules'
    user_options = distutils.command.build_py.build_py.user_options + [
        ('inplace',     'i', 'ignore build-lib and put compiled UI modules into the source '
                             'directory alongside your pure Python modules'),
    ]
    boolean_options = distutils.command.build_py.build_py.boolean_options + [
        'inplace']
    
    def initialize_options(self):
        distutils.command.build_py.build_py.initialize_options(self)
        self.qt_ui_files = None
        self.inplace = None
        
    def finalize_options(self):
        distutils.command.build_py.build_py.finalize_options(self)
        self.qt_ui_files = self.distribution.qt_ui_files
        self.set_undefined_options('build',
                                       ('inplace', 'inplace'),
                                   )
                
    def compile_ui(self, ui_file, py_file):
        from buildlib import translation_from_ui
        windowinfo = translation_from_ui.parse_file(ui_file)
        with open(py_file, 'wt', encoding='utf-8') as file:
            translation_from_ui.write_file(*windowinfo, file=file)
            
    def find_package_modules(self, package, package_dir):
        if package not in self.qt_ui_files:
            return distutils.command.build_py.build_py.find_package_modules(self, package, package_dir)
        modules = []
        for ui_file in self.qt_ui_files[package]:
            module = os.path.splitext(os.path.basename(ui_file))[0]
            modules.append((package, module, ui_file))
        return modules
        
    def build_module(self, module, infile, package):
        if infile and infile.endswith('.py'):
            return distutils.command.build_py.build_py.build_module(self, module, infile, package)
        if self.inplace:
            package_dir = self.get_package_dir(package)
            py_file = os.path.join(package_dir, module) + '.py'
        else:
            package = package.split('.')
            py_file = self.get_module_outfile(self.build_lib, package, module)
            self.mkpath(os.path.dirname(py_file))
        self.make_file(infile, py_file, self.compile_ui, [infile, py_file])

class build_models(Command):
    description = 'build data for Pybik models'
    user_options = [
        ('inplace',     'i', 'ignore build-lib and put compiled UI modules into the source '
                             'directory alongside your pure Python modules'),
        ('force',       'f', 'forcibly build everything (ignore file timestamps)'),
        ('parallel=',   None,'number of parallel build jobs or "True"'),
        ('fast',        None,'Build models without optimization'),
        ('pickle-protocol=', None, 'Pickle protocol for modeldata (default: highest protocol)'),
        ('maxsize=',    None, 'Maximum model size (default: 10)'),
        ('reproducible', None, 'Build modeldata reproducible (experimental)'),
        ('debug=', None, 'Comma-separated list of debug flags'),
    ]
    boolean_options = ['inplace', 'force', 'fast', 'reproducible']
    
    def initialize_options(self):
        self.build_base = None
        self.inplace = None
        self.force = None
        self.parallel = None
        self.fast = None
        self.pickle_protocol = None
        self.maxsize = None
        self.reproducible = None
        self.debug = None
        
    def finalize_options(self):
        self.set_undefined_options('build',
                                       ('build_base', 'build_base'),
                                       ('inplace', 'inplace'),
                                       ('force', 'force'),
                                       ('parallel', 'parallel'),
                                       ('fast', 'fast'),
                                       ('pickle_protocol', 'pickle_protocol'),
                                       ('maxsize', 'maxsize'),
                                       ('reproducible', 'reproducible'),
                                   )
        if isinstance(self.parallel, str):
            if self.parallel.lower() == 'true':
                self.parallel = True
            else:
                self.parallel = int(self.parallel)
        self.pickle_protocol = -1 if self.pickle_protocol is None else int(self.pickle_protocol)
        if self.maxsize is None:
            self.maxsize = 10
        else:
            from ast import literal_eval
            self.maxsize = literal_eval(self.maxsize)
        self.reproducible = False if self.reproducible is None else bool(self.reproducible)
        if self.debug is not None:
            import pybiklib.debug
            known_debug_flags = [d[6:].lower() for d in pybiklib.debug.__all__ if d.startswith('DEBUG_')]
            debug_flags = []
            for d in self.debug.split(','):
                if d not in known_debug_flags:
                    print('unknown debug option:', d)
                elif d:
                    debug_flags.append(d)
            pybiklib.debug.set_flags(debug_flags)
            print('debug flags:', ' '.join(debug_flags))
                
    def run(self):
        from buildlib import modeldata
        modeldir = os.path.join('data' if self.inplace else self.build_base, 'models')
        self.mkpath(modeldir)
        if self.force and not self.dry_run:
            for filename in os.listdir(modeldir):
                filename = os.path.join(modeldir, filename)
                os.remove(filename)
        if self.force:
            testfunc = None
        else:
            testfunc = lambda filename: (any(newer(os.path.join('buildlib', f), filename) for f in
                                                ('modeldata.py', 'modeldef.py', 'geom.py')))
        if testfunc is None or testfunc(modeldata.get_indexfilename(modeldir)):
            message = 'generating model data, this may take some time'
            def create_modeldata():
                modeldata.create_modeldata(modeldir, testfunc, self.parallel, self.fast,
                                        self.pickle_protocol, self.maxsize, self.reproducible)
            self.execute(create_modeldata, [], msg=message)
        data_files = self.distribution.data_files
        if self.dry_run:
            return
        for filename in os.listdir(modeldir):
            sourcepath = os.path.join(modeldir, filename)
            data_files.append(('share/pybik/models', (sourcepath,)))
        
        
class build_i18n(Command):
    description = 'compile message catalog to binary format'
    
    user_options = [
        ('inplace',         'i',  'ignore build-lib and put compiled UI modules into the source '
                                  'directory alongside your pure Python modules'),
        ('force',           'f',  'forcibly build everything (ignore file timestamps)'),
    ]
        
    boolean_options = ['inplace', 'force']
    
    def initialize_options(self):
        self.build_base = None
        self.inplace = None
        self.force = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('inplace', 'inplace'),
                                            ('force', 'force'),
                                  )
            
    def run(self):
        """Compile message catalog to binary format"""
        from buildlib.utils import po_isempty
        
        if self.inplace:
            mo_dir = 'data/locale'
        else:
            mo_dir = os.path.join(self.build_base, 'mo')
        data_files = self.distribution.data_files
        domain = self.distribution.metadata.name
        
        for po_file in glob('po/*.po'):
            if po_isempty(po_file, verbose=False):
                print('skipping empty po file', po_file)
                continue
            lang = os.path.splitext(os.path.basename(po_file))[0]
            mo_dir_lang =  os.path.join(mo_dir, lang, "LC_MESSAGES")
            mo_file = os.path.join(mo_dir_lang, "%s.mo" % domain)
            self.mkpath(mo_dir_lang)
            def msgfmt(po_file, mo_file):
                self.spawn(["msgfmt", po_file, "-o", mo_file])
            self.make_file([po_file], mo_file, msgfmt, [po_file, mo_file])
            
            targetpath = os.path.join("share/locale", lang, "LC_MESSAGES")
            data_files.append((targetpath, (mo_file,)))
            
            
class build_desktop(Command):
    description = 'build translated desktop files'
    
    user_options = [
        ('inplace',         'i',  'ignore build-lib and put compiled UI modules into the source '
                                  'directory alongside your pure Python modules'),
        ('force',           'f',  'forcibly build everything (ignore file timestamps)'),
    ]
        
    boolean_options = ['inplace', 'force']
    
    def initialize_options(self):
        self.build_base = None
        self.inplace = None
        self.force = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('inplace', 'inplace'),
                                            ('force', 'force'),
                                  )
        self.desktop_dir = os.path.join('data' if self.inplace else self.build_base, 'applications')
            
    def run(self):
        from buildlib import create_docs
        
        if self.distribution.desktop_templates is None:
            return
            
        data_files = self.distribution.data_files
        
        self.mkpath(self.desktop_dir)
        
        # merge .desktop.in with translation
        targetpath, templates = self.distribution.desktop_templates
        targetfiles = []
        for template in templates:
            templatefile = create_docs.get_template_filename(template)
            tmpfile = os.path.join(self.desktop_dir, os.path.basename(templatefile))
            dstfile = os.path.join(self.desktop_dir, os.path.basename(template))
            if not dstfile.endswith('.desktop'):
                self.warn('Skipping %s, seems not to be a desktop file template.' % templatefile)
                continue
            def make_translated_desktopfile(src, tmp, dst):
                create_docs.create_doc(src, tmp, skip='deb')
                self.spawn(["intltool-merge", '-d', 'po', tmp, dst])
                os.remove(tmp)
            self.make_file(glob('po/*.po') + [templatefile], dstfile,
                           make_translated_desktopfile, [template, tmpfile, dstfile])
            targetfiles.append(dstfile)
        data_files.append((targetpath, targetfiles))
        
        
class build_man (Command):
    description = 'build the manpage'
    user_options = [
        ('force',       'f',  'forcibly build everything (ignore file timestamps)'),    ]
    
    def initialize_options(self):
        self.force = None
        self.build_base = None
        
    def finalize_options(self):
        self.set_undefined_options('build', ('build_base', 'build_base'),
                                            ('force', 'force'),                                  )
        
    def run(self):
        def create_manpage(script_file, section, man_file):
            import re
            from buildlib.utils import modify_file
            os.environ['PYTHONPATH'] = '.'
            self.spawn(['help2man', '--name', config.SHORT_DESCRIPTION, '--section', section,
                                    '--output', man_file, '--no-info', script_file])
            modify_file(man_file, [
                    (r'^(?a)(\.TH\s+\S+\s+\S+\s+)".+?"(.*)$', r'\1"{}"\2'.format(re.escape(config.RELEASE_DATE)))])
            
        man_dir = os.path.join(self.build_base, 'man')
        self.mkpath(man_dir)
        if '.' not in sys.path:
            sys.path.insert(0, '.')
        section = '6'
        if self.dry_run:
            return
        for script_file in self.get_finalized_command('build_scripts').scripts:
            script, ext = os.path.splitext(os.path.basename(script_file))
            man_file = os.path.join(man_dir, '.'.join((script, section)))
            self.make_file(['pybiklib/config.py', script_file], man_file,
                           create_manpage, [script_file, section, man_file])
            self.distribution.data_files.append(('share/man/man'+section, [man_file]))
            
            
class build_doc (Command):
    description = "build documentation files"
    user_options = [
        ('inplace',         'i',  'ignore build-lib and put compiled UI modules into the source '
                                  'directory alongside your pure Python modules'),
    ]
    boolean_options = ['inplace']
    
    def initialize_options(self):
        self.inplace = None
        self.build_base = None
        self.doc_files = []
        
    def finalize_options(self):
        self.set_undefined_options('build', ('inplace', 'inplace'), ('build_base', 'build_base'))
        
    def run(self):
        from buildlib import create_docs
        
        info('generating documentation files')
        self.mkpath(os.path.join(self.build_base, 'doc'))
        for basename, skip in [
                    ('copyright', 'deb'),
                    ('README', None),
                    ('INSTALL', None),
                ]:
            filename = None if self.inplace else os.path.join(self.build_base, 'doc', basename)
            info('generating file: %r', filename or basename)
            create_docs.create_doc(basename, filename, skip=skip)
            if filename:
                self.doc_files.append(filename)
        
        
class install_lib (distutils.command.install_lib.install_lib):
    user_options = distutils.command.install_lib.install_lib.user_options + [
            ('arch-only', None, 'Install only architecture dependent files'),
            ('indep-only', None, 'Install only architecture independent files'),
        ]
    boolean_options = ['arch-only', 'indep-only']
    
    def initialize_options(self):
        distutils.command.install_lib.install_lib.initialize_options(self)
        self.data_dir = None
        self.arch_only = None
        self.indep_only = None
        
    def finalize_options(self):
        distutils.command.install_lib.install_lib.finalize_options(self)
        self.set_undefined_options('install',
                                       ('arch_only', 'arch_only'),
                                       ('indep_only', 'indep_only'),
                                   )
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
            
    def build(self):
        if not self.skip_build:
            if self.distribution.has_pure_modules() and self.indep_only:
                self.run_command('build_py')
            if self.distribution.has_ext_modules() and self.arch_only:
                self.run_command('build_ext')
                
                
class install_desktop(Command):
    description = 'install desktop files'
    user_options = [('ask', None, 'Ask before installation'),]
    boolean_options = ['ask']
    
    def initialize_options(self):
        self.ask = False
        
    def finalize_options(self):
        pass
    
    def run(self):
        if self.ask:
            try:
                answer = input('\nInstall desktop file? [Yn]: ')
            except EOFError:
                print()
                return
            if answer.lower() not in ['y', '']:
                return
                
        self.run_command('build_desktop')
        desktop_dir = self.get_finalized_command('build_desktop').desktop_dir
        target_dir = os.path.join(config.get_data_home(), 'applications')
        
        def keyabspath(line, key, file=None):
            key += '='
            if line.startswith(key):
                if file is None:
                    file = line.split('=', 1)[1].rstrip()
                line = key + os.path.abspath(file) + '\n'
            return line
            
        self.mkpath(target_dir)
        for fnin in glob('%s/*.desktop' % desktop_dir):
            fnout = os.path.join(target_dir, os.path.basename(fnin))
            info('installing desktop file to: %s', fnout)
            with open(fnin, 'rt', encoding='utf-8') as fin, open(fnout, 'wt', encoding='utf-8') as fout:
                for line in fin:
                    line = keyabspath(line, 'TryExec')
                    line = keyabspath(line, 'Exec')
                    line = keyabspath(line, 'Icon', config.APPICON192_FILE)
                    fout.write(line)
                    
                    
class install (install_orig):
    user_options = install_orig.user_options + [
            ('data-dir=', 't', 'Directory where the application will find the data'),
            ('arch-only', None, 'Install only architecture dependent files'),
            ('indep-only', None, 'Install only architecture independent files'),
        ]
    boolean_options = ['arch-only', 'indep-only']
        
    def initialize_options(self):
        install_orig.initialize_options(self)
        self.data_dir = None
        self.arch_only = False
        self.indep_only = False
        
    def finalize_options(self):
        install_orig.finalize_options(self)
        if self.data_dir is None:
            self.data_dir = os.path.join(self.install_data, 'share')
        if not self.arch_only and not self.indep_only:
            self.arch_only = self.indep_only = True
        if not self.indep_only:
            self.__class__.sub_commands = [(cmd, func)
                            for cmd, func in install_orig.sub_commands
                                if cmd == 'install_lib']
            
    def run(self):
        if not self.skip_build:
            # install_orig.run() will run build, but we need
            # to modify a file between build and install
            build_cmd = self.distribution.get_command_obj('build')
            build_cmd.arch_only = self.arch_only
            build_cmd.indep_only = self.indep_only
            self.run_command('build')
            self.skip_build = True
            
        filename = os.path.join(self.build_lib, 'pybiklib', 'config.py')
        app_data_dir = os.path.join(self.data_dir, 'pybik')
        if self.indep_only:
            from buildlib.utils import modify_file
            modify_file(filename, [
                    (r'^(data_dir\s*=\s*).*$',      r'\1' + repr(self.data_dir)),
                    (r'^(appdata_dir\s*=\s*).*$',   r'\1' + repr(app_data_dir)),])
        install_orig.run(self)
        
        
class clean (distutils.command.clean.clean):
    user_options = distutils.command.clean.clean.user_options + [
            ('inplace', 'i',         'clean up files in the source directory'),
        ]
    boolean_options = distutils.command.clean.clean.boolean_options + [
            'inplace',
        ]
        
    def initialize_options(self):
        distutils.command.clean.clean.initialize_options(self)
        self.inplace = None
        
    def finalize_options(self):
        distutils.command.clean.clean.finalize_options(self)
        self.set_undefined_options('build',
                                       ('inplace', 'inplace'),
                                   )
        
    def run(self):
        def remove_tree_(directory):
            if os.path.exists(directory):
                remove_tree(directory, dry_run=self.dry_run, verbose=self.verbose)
            else:
                warn("%r does not exist -- can't clean it", directory)
        def remove_file(filename):
            if os.path.exists(filename):
                info('removing %r', filename)
                if not self.dry_run:
                    os.remove(filename)
            else:
                warn("%r does not exist -- can't clean it", filename)
                
        if self.all:
            for _dir in ['applications', 'doc', 'man', 'mo', 'models']:
                remove_tree_(os.path.join(self.build_base, _dir))
        distutils.command.clean.clean.run(self)
        remove_file('MANIFEST')
        if self.inplace:
            for dirname in ('buildlib/', 'pybiklib/', 'pybiklib/plugins/', 'pybiklib/ui/', 'pybiktest/', 'tools/'):
                remove_tree_(dirname + '__pycache__')
            if self.all:
                dirnames = ['data/applications', 'data/locale', 'data/models']
                for dirname in dirnames:
                    remove_tree_(dirname)
                for filename in glob('pybiklib/*.so'):
                    remove_file(filename)
                for filename in glob('pybiklib/ui/*.py'):
                    if filename != 'pybiklib/ui/__init__.py':
                        remove_file(filename)
                for filename in ['copyright', 'INSTALL', 'README', 'pybik']:
                    remove_file(filename)
                
                
class sdist(distutils.command.sdist.sdist):
    user_options = distutils.command.sdist.sdist.user_options + [
            ('debian-names', None,  'Create archive files with Debian names'),
        ]
    boolean_options = distutils.command.sdist.sdist.boolean_options + [
            'debian-names'
        ]
        
    def initialize_options(self):
        distutils.command.sdist.sdist.initialize_options(self)
        self.debian_names = False
        self.desktop_dir = None
        
    def finalize_options(self):
        distutils.command.sdist.sdist.finalize_options(self)
        self.set_undefined_options('build_desktop', ('desktop_dir', 'desktop_dir'))
        
    def run(self):
        self.run_command('build_doc')
        self.run_command('build_csrc')
        self.run_command('build_desktop')
        distutils.command.sdist.sdist.run(self)
        
        for archive_file in self.archive_files:
            if self.debian_names:
                debian_file = archive_file.replace('-', '_', 1).replace('.tar', '.orig.tar', 1)
                os.rename(archive_file, debian_file)
        
    def get_file_list(self):
        from buildlib.utils import po_isempty
        for f in glob('po/*.po'):
            if po_isempty(f):
                print('skipping empty po file', f)
            else:
                self.filelist.files.append(f)
        distutils.command.sdist.sdist.get_file_list(self)
        
    def make_release_tree(self, base_dir, files):
        distutils.command.sdist.sdist.make_release_tree(self, base_dir, files)
        # Release date
        from datetime import date
        from buildlib.utils import modify_file
        modify_file(os.path.join(base_dir, 'pybiklib', 'config.py'),[
                (r'^(RELEASE_DATE\s*=\s*).*$',  r'\1' + repr(date.today().isoformat())),
            ])
        # doc files
        for filename in self.get_finalized_command('build_doc').doc_files:
            self.copy_file(filename, base_dir)
        # csrc
        dst_dir = os.path.join(base_dir, 'csrc')
        self.mkpath(dst_dir)
        extensions = self.get_finalized_command('build_csrc').extensions
        for extension in extensions:
            for src in extension.sources:
                self.copy_file(src, dst_dir)
        # desktop files
        dst_dir = os.path.join(base_dir, 'data', 'applications')
        self.mkpath(dst_dir)
        for template in self.distribution.desktop_templates[1]:
            src = os.path.join(self.desktop_dir, template)
            self.copy_file(src, dst_dir)
        
        
setup(
        # Metadata
        name=config.PACKAGE,
        version=config.VERSION,
        author=config.AUTHOR,
        author_email=config.CONTACT_EMAIL,
        url=config.WEBSITE,
        download_url=config.DOWNLOADS,
        
        description=config.SHORT_DESCRIPTION,
        long_description='\n'.join(config.LONG_DESCRIPTION),
        classifiers=[
                'Development Status :: 5 - Production/Stable',
                'Environment :: X11 Applications :: Qt',
                'Intended Audience :: End Users/Desktop',
                'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
                'Operating System :: POSIX',
                'Operating System :: POSIX :: Linux',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.4',
                'Programming Language :: Cython',
                #TODO: 'Programming Language :: OpenGL Shading Language',
                'Topic :: Games/Entertainment :: Puzzle Games',
            ],
        keywords='rubik cube puzzle game',
        license=config.LICENSE_NAME,
        # Metadata added in overloaded Distribution
        bug_contact=config.CONTACT_FILEBUG,
        
        # Files
        scripts=['pybiklib/pybik.py'],
        data_files=[
                        ('share/pixmaps', ['data/icons/pybik.png']),
                        ('share/pybik/icons', ['data/icons/pybik-64.png']),
                        ('share/pybik/icons', ['data/icons/pybik-192.png']),
                        ('share/pybik/ui/cursors', glob('data/ui/cursors/*')),
                        ('share/pybik/ui/images', glob('data/ui/images/*')),
                        ('share/pybik/ui/qt', glob('data/ui/qt/*')),
                        ('share/pybik/plugins', glob('data/plugins/*')),
                        ('share/pybik/shaders', glob('data/shaders/*')),
                        ('share/pybik/tests', glob('data/tests/*')),
                        ('share/pybik/', ['data/GPL-3']),
                    ],
        desktop_templates=('share/applications', ['pybik.desktop']),
        ext_modules=[
                        Extension('pybiklib/_gldraw',
                                    ['pybiklib/gldraw.py'],
                                    depends=['pybiklib/gl.pxd'],
                                    libraries=['GL'],
                                    define_macros=[('GL_GLEXT_PROTOTYPES', None)]),
                        Extension('pybiklib/_glarea',
                                    ["pybiklib/glarea.py"],
                                    depends=['pybiklib/gl.pxd',
                                             'pybiklib/gldraw.py'],
                                    libraries=['GL'],
                                    define_macros=[('GL_GLEXT_PROTOTYPES', None)]),
                    ],
        packages=['pybiklib', 'pybiklib.plugins', 'pybiklib.ui', 'pybiktest'],
        qt_ui_files={'pybiklib.ui': glob('data/ui/qt/*.ui')+['pybiklib/ui/__init__.py']},
        
        # setup classes
        distclass=Distribution,
        cmdclass={
                    'build': build,
                    'build_csrc': build_csrc,
                    'build_ext': build_ext,
                    'build_scripts': build_scripts,
                    'build_py': build_py,
                    'build_models': build_models,
                    'build_i18n': build_i18n,
                    'build_desktop': build_desktop,
                    'build_man': build_man,
                    'build_doc': build_doc,
                    'install_data': distutils.command.install_data.install_data,
                    'install_lib': install_lib,
                    'install_desktop': install_desktop,
                    'install': install,
                    'clean': clean,
                    'sdist': sdist,
                },
     )
     



# -*- coding: utf-8 -*-

#  Copyright © 2014  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import deque


class PatternCubie (list):
    # value: (index, rotation)
    def __init__(self, model, position=None):
        self.model = model
        self.position = position
        
    def gray_conditions(self):
        for idx in range(self.model.cell_count):
            for rot in self.model.face_permutations.keys():
                pos = self.model.rotated_position[rot][idx]
                if pos == self.position:
                    yield idx, rot
                    
    def set_gray(self):
        self[:] = list(self.gray_conditions())
        return self
        
    def parse(self, position, condition):
        self.parse_or_condition(position, condition)
        assert self.position is not None, (self.position, position, condition)
        return self
        
    def __str__(self):
        pos = ''.join(self.model.cells_visible_faces[self.position])
        vals = []
        notconditions = list(self.not_conditions(self))
        if len(notconditions) == 1:
            conditions = notconditions
            prefix = '!'
        else:
            conditions = self
            prefix = ''
        grouped = {}
        for idx, rot in conditions:
            grouped.setdefault(idx, []).append((idx, rot))
        def cond2str(idx, rot):
            return ''.join(self.model.face_symbolic_to_face_color(face, rot) for face in pos).lower()
        for idx, conds in grouped.items():
            if len(conds) > 1 and set(conds) == set(self.rotated_conditions(idx)):
                grouped[idx] = '*' + cond2str(*conds[0])
            else:
                grouped[idx] = '|'.join(cond2str(*c) for c in conds)
        return '{}={}{}'.format(pos.lower(), prefix, '|'.join(grouped.values()))
        
    def parse_pattern_condition(self, position, condition, order=True):
        if order:
            def innergen():
                for pface, cface in zip(position, condition):
                    if '?' != cface != self.model.face_symbolic_to_face_color(pface, rot):
                        break
                else:
                    yield idx, rot
        else:
            def innergen():
                cond = condition
                for pface in position:
                    cface = self.model.face_symbolic_to_face_color(pface, rot)
                    if cface in cond:
                        cond = cond.replace(cface, '', 1)
                    elif '?' in cond:
                        cond = cond.replace('?', '', 1)
                    else:
                        break
                else:
                    yield idx, rot
        position = position.upper()
        condition = condition.upper()
        for idx in range(self.model.cell_count):
            for rot in self.model.face_permutations.keys():
                pos = self.model.rotated_position[rot][idx]
                if set(self.model.cells_visible_faces[pos]) == set(position):
                    if self.position is None:
                        self.position = pos
                    else:
                        assert self.position == pos, (self.position, pos)
                    yield from innergen()
                    
    def not_conditions(self, conditions):
        for c in self.gray_conditions():
            if c not in conditions:
                yield c
                
    def rotated_conditions(self, idx):
        for rot in self.model.face_permutations.keys():
            pos = self.model.rotated_position[rot][idx]
            if pos == self.position:
                yield idx, rot
                
    def parse_prefix_condition(self, position, condition):
        if condition.startswith('!'):
            if condition.startswith('!*'):
                conditions = [c for c in self.parse_pattern_condition(position, condition[2:], False)]
            else:
                conditions = [c for c in self.parse_pattern_condition(position, condition[1:], True)]
            citer = self.not_conditions(conditions)
        elif condition.startswith('*'):
            #TODO: Instead of testing only rotated conditions, test all permutations.
            #      This should not break existing rules, and would allow to match
            #      e.g. dfr and dfl. Could be done by comparing sorted strings after
            #      expanding patterns.
            #DONE: use this in plugins module
            citer = self.parse_pattern_condition(position, condition[1:], False)
        else:
            citer = self.parse_pattern_condition(position, condition, True)
        for c in citer:
            self.union(*c)
                
    def parse_or_condition(self, position, condition):
        for c in condition.split('|'):
            self.parse_prefix_condition(position, c)
            
    def copy(self):
        cubie = self.__class__(self.model, self.position)
        cubie.extend(list(self))
        return cubie
        
    def count(self):
        return len(self)
        
    def union(self, idx, rot):
        self.append((idx, rot))
        
    def fork(self, condition):
        assert self.position == condition.position, (self.position, condition.position)
        assert self
        accepts, rejects = self.__class__(self.model, self.position), self.__class__(self.model, self.position)
        for c in self:
            if c in condition:
                accepts.append(c)
            else:
                rejects.append(c)
        return accepts or None, rejects or None
        
    def identify_rotation_blocks(self, maxis, mslice):
        return mslice == -1 or mslice == self.model.cell_indices[self.position][maxis]
        
    def transform(self, move):
        assert self
        axis, slice_, dir_ = move.data
        pos = self.position
        newpos = None
        irs = []
        if self.identify_rotation_blocks(axis, slice_):
            for idx, rot in self:
                newpos_, rot = self.model.rotate_symbolic(axis, dir_, pos, rot)
                if newpos is None:
                    newpos = newpos_
                else:
                    assert newpos == newpos_
                irs.append((idx, rot))
            assert newpos is not None
            cubie = self.__class__(self.model, newpos)
            cubie.extend(irs)
            assert cubie
            return cubie
        else:
            return self
            
    def check(self):
        if not self:
            warn('cubie is empty')
        if len(self) != len(set(self)):
            assert False, 'cubie has duplicates'
        if self == list(self.gray_conditions()):
            assert False, 'cubie is gray'
        
        
class PatternCube (dict):
    # key: position
    # value: list of (index, rotation)
    def __init__(self, model):
        self.model = model
        # empty list matches everything, cubies added restrict matches
        # a PatternCube object that never matches is invalid, use None instead
        
    def parse(self, conditions):
        self.parse_and_condition(conditions)
        return self
        
    def parse_and_condition(self, conditions):
        for pos, cond in conditions:
            cubie = PatternCubie(self.model).parse(pos, cond)
            assert cubie
            self[cubie.position] = cubie
            
    def __str__(self):
        return ' '.join(str(v) for v in self.values())
        
    def __bool__(self):
        return True
        
    def copy(self):
        cube = self.__class__(self.model)
        for k, v in self.items():
            cube[k] = v.copy()
        return cube
        
    def count(self):
        cnt = 1
        for pos in range(self.model.cell_count):
            if pos in self:
                cnt *= self[pos].count()
            else:
                cnt *= PatternCubie(self.model, pos).set_gray().count()
        return cnt
        
    def fork(self, condition):
        accepts = self.copy()
        rejects = []
        for pos, cond in condition.items():
            # indices in self but not in condition can be ignored
            if pos in self:
                cubie = self[pos]
            else:
                cubie = PatternCubie(self.model, pos).set_gray()
            accept, reject = cubie.fork(cond)
            if accept:
                if reject:
                    accepts[pos] = accept
                    reject_ = self.copy()
                    reject_[pos] = reject
                    reject_.check()
                    rejects.append(reject_)
                # else: accept is full cubie, nothing to do
            else:
                if reject:
                    self.check()
                    return None, [self.copy()]
                else:
                    assert False, 'empty PatternCube not allowed'
        return accepts, rejects
        
    def transform(self, move):
        cube = self.__class__(self.model)
        for pos, cubie in list(self.items()):
            cubie = cubie.transform(move)
            cube[cubie.position] = cubie
        return cube
        
    def check(self):
        for pos, cubie in self.items():
            assert pos == cubie.position
            cubie.check()
            
    def check_empty_simple(self):
        for pos1, cubie1 in self.items():
            if len(cubie1) == 1:
                idx1, rot1 = cubie1[0]
                for pos2, cubie2 in self.items():
                    if pos1 != pos2:
                        if [idx2 for idx2, rot2 in cubie2 if idx1 == idx2]:
                            cubie2[:] = [(idx2, rot2) for idx2, rot2 in cubie2 if idx1 != idx2]
                            if not cubie2:
                                return True
        return False
        
    def check_empty_full(self):
        indexlist = []
        for cubie in self.values():
            indexlist.append(set(idx for idx, rot in cubie))
        iters = deque()
        sample = deque()
        i = 0
        while i < len(indexlist):
            # new iter for i, must have at least one item
            iit = iter(indexlist[i])
            try:
                iidx = next(iit)
            except StopIteration:
                return True
            # idx must be unique
            forceloop = False
            while forceloop or iidx in sample:
                try:
                    iidx = next(iit)
                    forceloop = False
                except StopIteration:
                    # no unique idx, continue with prev iter
                    try:
                        iit = iters.pop()
                    except IndexError:
                        return True
                    sample.pop()
                    i -= 1
                    forceloop = True
            iters.append(iit)
            sample.append(iidx)
            i += 1
        return False
        
    def join(self, cube):
        issubset = True
        issuperset = True
        diffs = []
        for pos in range(self.model.cell_count):
            if pos in self:
                selfv = set(self[pos])
                if pos in cube:
                    cubev = set(cube[pos])
                    issubset = issubset and selfv.issubset(cubev)
                    issuperset = issuperset and selfv.issuperset(cubev)
                    if selfv != cubev:
                        diffs.append(pos)
                else:
                    issuperset = False
                    diffs.append(pos)
            else:
                if pos in cube:
                    issubset = False
                    diffs.append(pos)
                else:
                    pass
        if issubset:
            return cube
        if issuperset:
            return self
        if len(diffs) == 1:
            res = self.copy()
            pos = diffs[0]
            res[pos].clear()
            res[pos].extend(list(set(self[pos] + cube[pos])))
            assert res[pos]
            return res
        return None
            
    def rotate_colors(self, rot):
        perm = self.model.rotated_position[rot]
        cube = PatternCube(self.model)
        for pos, cubie in self.items():
            assert cubie
            permpos = perm.index(pos)
            newcubie = PatternCubie(self.model, pos)
            for cidx, crot in cubie:
                ridx = perm.index(cidx)
                rrot = self.model.norm_symbol(rot + crot)
                newcubie.union(ridx, rrot)
            assert newcubie
            cube[pos] = newcubie
        return cube
        
        




#-*- coding:utf-8 -*-

#  Copyright © 2009-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
from collections import namedtuple
import math
from time import sleep
from contextlib import contextmanager

#XXX: Workaround for lp:941826, "dlopen(libGL.so) resolves to mesa rather than nvidia"
#     Was: "PyQt cannot compile shaders with Ubuntu's Nvidia drivers"
#     https://bugs.launchpad.net/ubuntu/+source/python-qt4/+bug/941826
import ctypes.util
ctypes.CDLL(ctypes.util.find_library('GL'), ctypes.RTLD_GLOBAL)

from PyQt5.QtCore import QElapsedTimer, QMutex, Qt, QThread, QTimer, QWaitCondition
from PyQt5.QtCore import pyqtSignal as Signal
from PyQt5.QtGui import (QColor, QCursor, QImage, QPixmap, QTransform, QSurfaceFormat,
                        QOpenGLTexture, QOpenGLFramebufferObject)
from PyQt5.QtWidgets import QOpenGLWidget

from .debug import (debug, DEBUG, DEBUG_MSG, DEBUG_DRAW, DEBUG_MOUSEPOS, DEBUG_FPS, DEBUG_VFPS, 
                    DEBUG_PUREPYTHON, DEBUG_LIVESHADER, DEBUG_LOGGL)

if DEBUG_LOGGL:
    try:
        import OpenGL
        OpenGL.FULL_LOGGING = True
        import OpenGL.GL
    except ImportError:
        pass

from . import config
from .theme import Theme
from .settings import settings
from .model import Model


class ShaderInterface:
    shader_prefix = b'''
#version 120
#line 0
'''
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.shadername = None
        
    def gl_init_shaders(self):
        self.gl_load_shader()
        if DEBUG_DRAW:
            debug('Creating "hud" shaders:')
            vertex_source = self._read_shader('hud.vert')
            fragment_source = self._read_shader('pick.frag')
            self.glarea.gl_create_hud_program(vertex_source, fragment_source)
        debug('Creating "pick" shaders:')
        vertex_source = self._read_shader('pick.vert')
        fragment_source = self._read_shader('pick.frag')
        self.glarea.gl_create_pick_program(vertex_source, fragment_source)
        
    @staticmethod
    def _read_shader(filename):
        if DEBUG_MSG or DEBUG_LIVESHADER:
            print('Loading shader:', filename)
        filename = os.path.join(config.SHADER_DIR, filename)
        with open(filename, 'rb') as sfile:
            source = sfile.read()
        return ShaderInterface.shader_prefix + source
        
    def gl_set_shader(self):
        if self.shadername is None:
            return
        vertex_source = self._read_shader(self.shadername + '.vert')
        fragment_source = self._read_shader(self.shadername + '.frag')
        self.glarea.gl_create_render_program(vertex_source, fragment_source)
        
    def gl_load_shader(self):
        self.shadername = settings.draw.shader_nick
        self.gl_set_shader()
        
        
class AnimationInterface:
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        self.stop_requested = False
        self.animation_active = False
        
        self.timer_animate = QTimer()
        self.timer_animate.timeout.connect(self._on_animate)
        
    def animate_rotation(self, move_data, blocks, stop_after):
        self.stop_requested = stop_after
        angle = 360. / self.model.symmetries[move_data.axis]
        axisx, axisy, axisz = self.model.axes[move_data.axis]
        if move_data.dir:
            self.gldraw.set_animation_start(blocks, angle, -axisx, -axisy, -axisz)
        else:
            self.gldraw.set_animation_start(blocks, angle, axisx, axisy, axisz)
        self.animation_active = True
        self.timer_animate.start(0 if DEBUG_VFPS else 20)
        self.update_selection()
            
    def animation_end(self):
        self.animation_active = False
        self.timer_animate.stop()
        
    def _on_animate(self):
        increment = self.speed * 1e-02 * 20
        increment = min(increment, 45)
        if self.gldraw.set_animation_next(increment):
            self.update()
            return
            
        self.animation_ending.emit(self.stop_requested)
        self.update()
        self.update_selection()
        self.stop_requested = False
        
    def animate_abort(self, update=True):
        if update:
            self.animation_ending.emit(self.stop_requested)
            self.update()
            self.update_selection()
        else:
            self.animation_end()
        self.stop_requested = False
        
        
class CubeArea (QOpenGLWidget, ShaderInterface, AnimationInterface):
    animation_ending = Signal(bool)
    request_rotation = Signal((int, int, bool))
    request_swap_blocks = Signal((int, int, int, bool))
    request_rotate_block = Signal((int, bool))
    drop_color = Signal((int, str, str))
    drop_file = Signal((int, str, str))
    debug_info = Signal(str, object)
    
    def __init__(self, opts, model, **kwargs):
        super().__init__(**kwargs)
        
        self.texture = None
        
        self.theme = Theme()
        self.mirror_distance = None
        self.model = model
        
        self.last_mouse_x = -1
        self.last_mouse_y = -1
        self.button_down_background = False
        self.mouse_xy = -1, -1
        self.pickdata = None
        self.editing_model = False
        settings.keystore.changed.connect(self.on_settings_changed, Qt.QueuedConnection)
        
        if DEBUG_FPS:
            self.monotonic_time = QElapsedTimer()
            self.monotonic_time.start()
            self.render_count = 0
            self.fps = 0.
            
        self.speed = settings.draw.speed
        
        # Initialize the render engine
        self.gldraw, self.glarea = self.import_gl_engine()
        self.glarea.init_module()
        self.gldraw.init_module()
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(*self.model.default_rotation)
        self.glarea.set_antialiasing(self.format().samples() > 1)
        self.init_theme()
        
        self.setAcceptDrops(True)
        self.setFocusPolicy(Qt.StrongFocus)
        self.setMinimumSize(300, 300)
        
        self.cursors = None
        self.load_cursors()
        self.update_selection_pending = False
        self.timer_update_selection = QTimer(self)
        self.timer_update_selection.setSingleShot(True)
        self.timer_update_selection.timeout.connect(self.on_idle_update_selection)
        self.set_cursor()
        self.setMouseTracking(True)
        
    @classmethod
    def set_default_surface_format(cls):
        glformat = QSurfaceFormat()
        if settings.draw.samples > 0:
            glformat.setSamples(2**settings.draw.samples)
        if DEBUG_VFPS:
            glformat.setSwapInterval(0)
        cls.default_format = QSurfaceFormat.defaultFormat()
        QSurfaceFormat.setDefaultFormat(glformat)
        
    def import_gl_engine(self):
        if not DEBUG_PUREPYTHON:
            from . import _gldraw as gldraw
            from . import _glarea as glarea
        else:
            try:
                import OpenGL.GL
            except ImportError as e:
                print('The pure Python mode needs PyOpenGL (for Python 3):', e)
                raise SystemExit(1)
            from . import gldraw
            from . import glarea
        return gldraw, glarea
        
    @contextmanager
    def lock_glcontext(self):
        self.makeCurrent()
        yield
        self.doneCurrent()
        
    def stop(self):
        with self.lock_glcontext():
            self.glarea.gl_exit()
            self.gl_delete_textures()
        # this line prevents a segmentation fault with PySide if game->quit selected
        self.setMouseTracking(False)
        settings.keystore.changed.disconnect(self.on_settings_changed)
        self.timer_update_selection.timeout.disconnect(self.on_idle_update_selection)
        self.timer_animate.timeout.disconnect(self._on_animate)
        
    def load_cursors(self):
        cursors = []
        # Load 3 cursors from file (n - ne)
        for i, (x, y) in enumerate([(8, 0), (15, 0), (15, 0)]):
            filename = os.path.join(config.UI_DIR, 'cursors', 'mouse_{}.png'.format(i))
            image = QImage(filename)
            cursors.append((image, x, y))
        # 1 cursor (nnw)
        image, x, y = cursors[1]
        cursors.insert(0, (image.mirrored(True, False), 15-x, y))
        # 12 cursors (ene - nw)
        transform = QTransform()
        transform.rotate(90)
        for i in range(4, 16):
            image, x, y = cursors[-4]
            cursors.append((image.transformed(transform), 15-y, x))
        cursors.append(cursors[0])
        self.cursors = [QCursor(QPixmap.fromImage(image), x, y) for image, x, y in cursors[1:]]
        # cursor for center faces
        filename = os.path.join(config.UI_DIR, 'cursors', 'mouse_ccw.png')
        cursor = QCursor(QPixmap(filename), 7, 7)
        self.cursors.append(cursor)
        # background cursor
        cursor = QCursor()
        cursor.setShape(Qt.CrossCursor)
        self.cursors.append(cursor)
        
    def init_theme(self):
        rgba = QColor()
        for facekey, unused_facename in Model.cache_index['facenames']:
            rgba.setNamedColor(settings.theme.faces[facekey].color)
            self.theme.faces[facekey].color = rgba.red(), rgba.green(), rgba.blue()
            self.theme.faces[facekey].imagemode = settings.theme.faces[facekey].mode
        self.set_background_color(settings.theme.bgcolor)
        
    def load_face_texture(self, facekey):
        filename = settings.theme.faces[facekey].image
        if not self.theme.load_face_image(facekey, filename):
            del settings.theme.faces[facekey].image
        
    def gl_set_texture_atlas(self):
        width, height = 0, 0
        texrects = []
        faces = [self.theme.faces[fk] for fk, fn in Model.cache_index['facenames']]
        for fd in faces:
            x = width
            w = fd.image.width()
            h = fd.image.height()
            width += w
            height = max(h, height)
            texrects.append((x, w, h))
        w = h = 64
        while w < width:
            w *= 2
        while h < height:
            h *= 2
        width, height = w, h
        
        if self.texture is not None:
            self.texture.destroy()
        self.texture = QOpenGLTexture(QOpenGLTexture.Target2D)
        self.texture.setFormat(QOpenGLTexture.RGBA32F)
        self.texture.setSize(width, height)
        self.texture.setMinMagFilters(QOpenGLTexture.Linear, QOpenGLTexture.Linear)
        self.texture.allocateStorage()
        self.texture.bind()
        for (x, w, h), fd in zip(texrects, faces):
            fd.texrect = (x+.5)/width, .5/height, (x+w-.5)/width, (h-.5)/height
            data = fd.image.bits()
            data.setsize(fd.image.byteCount())
            data = bytes(data)
            self.gldraw.gl_set_atlas_texture(x, w, h, data)
            
    def gl_delete_textures(self):
        if self.texture is not None:
            self.texture.release()
            self.texture.destroy()
        for fd in self.theme.faces.values():
            fd.image = None
        
    def set_glmodel_full(self, model=None, rotations=None):
        if model is not None:
            self.model = model
            self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(*self.model.default_rotation)
        #TODO: The bounding_sphere_radius is optimised for the far clipping plane,
        #      for the near clipping plane the radius without the mirror_distance
        #      would be sufficient.
        s = max(self.model.sizes) if self.model.sizes else 1
        if settings.draw.mirror_faces:
            self.mirror_distance = settings.draw.mirror_distance * s
            sm = s + self.mirror_distance
        else:
            self.mirror_distance = None
            sm = s
        bounding_sphere_radius = math.sqrt(2*s*s + sm*sm)
        self.glarea.set_frustum(bounding_sphere_radius, settings.draw.zoom)
        
        self.set_glmodel_selection_mode(settings.draw.selection)
        if rotations is not None:
            self.set_transformations(rotations)
        self.update_selection()
        self.update()
        
    def set_glmodel_selection_mode(self, selection_mode):
        glmodeldata, self.glpickdata = self.model.gl_vertex_data(selection_mode, self.theme, self.mirror_distance)
        with self.lock_glcontext():
            self.gldraw.gl_set_data(*glmodeldata)
        
    def set_transformations(self, rotations):
        indices = self.model.rotations_symbolic_to_index(rotations)
        self.gldraw.set_transformations(indices)
        
    def gl_create_pickbuffer(self, width, height):
        self.pickbuffer = QOpenGLFramebufferObject(width, height)
        self.pickbuffer.setAttachment(QOpenGLFramebufferObject.Depth)
        
    def initializeGL(self):
        if DEBUG_MSG:
            glformat = self.format()
            glrformat = QSurfaceFormat.defaultFormat() # requested format
            gldformat = self.default_format
            def printglattr(name, *attrs):
                print('  {}: '.format(name), end='')
                def get_value(glformat, attr):
                    if isinstance(attr, str):
                        return getattr(glformat, attr)()
                    else:
                        return attr(glformat)
                values = [get_value(glformat, a) for a in attrs]
                rvalues = [get_value(glrformat, a) for a in attrs]
                dvalues = [get_value(gldformat, a) for a in attrs]
                print(*values, end='')
                if values != rvalues:
                    print(' (', end='')
                    print(*rvalues, end=')')
                if rvalues != dvalues:
                    print(' [', end='')
                    print(*dvalues, end=']')
                print()
            print('Surface format (requested (), default []):')
            printglattr('alpha', 'alphaBufferSize')
            printglattr('rgb', 'redBufferSize', 'greenBufferSize', 'blueBufferSize')
            printglattr('depth', 'depthBufferSize')
            printglattr('options', lambda glformat: str(int(glformat.options())))
            printglattr('profile', 'profile')
            printglattr('renderableType', 'renderableType')
            printglattr('samples', 'samples')
            printglattr('stencil', 'stencilBufferSize')
            printglattr('stereo', 'stereo')
            printglattr('swapBehavior', 'swapBehavior')
            printglattr('swapInterval', 'swapInterval')
            printglattr('version', lambda glformat: '{}.{}'.format(*glformat.version()))
        self.gl_init_shaders()
        for facekey, unused_facename in Model.cache_index['facenames']:
            self.load_face_texture(facekey)
        self.gl_set_texture_atlas()
        self.glarea.gl_init()
        self.gl_create_pickbuffer(self.width(), self.height())
        
    def paintGL(self):
        self.texture.bind()
        self.glarea.gl_render()
        if DEBUG:
            if DEBUG_DRAW:
                if self.pickdata is not None and self.pickdata.angle is not None:
                    self.glarea.gl_render_select_debug()
            if DEBUG_FPS:
                self.render_count += 1
                if self.monotonic_time.elapsed() > 1000:
                    elapsed = self.monotonic_time.restart()
                    self.fps = 1000. / elapsed * self.render_count
                    self.render_count = 0
                    self.debug_info.emit('fps', self.fps)
        
    def resizeGL(self, width, height):
        self.glarea.gl_resize(width, height)
        self.gl_create_pickbuffer(width, height)
        
    MODIFIER_MASK = int(Qt.ShiftModifier | Qt.ControlModifier)
    def keyPressEvent(self, event):
        modifiers = int(event.modifiers()) & self.MODIFIER_MASK
        if modifiers:
            return QOpenGLWidget.keyPressEvent(self, event)
        elif event.key() == Qt.Key_Right:
            self.rotation_x += 2
        elif event.key() == Qt.Key_Left:
            self.rotation_x -= 2
        elif event.key() == Qt.Key_Up:
            self.rotation_y -= 2
        elif event.key() == Qt.Key_Down:
            self.rotation_y += 2
        else:
            return QOpenGLWidget.keyPressEvent(self, event)
            
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(self.rotation_x, self.rotation_y)
        self.update()
        self.update_selection()
        
    PickData = namedtuple('PickData', 'maxis mslice mdir blockpos symbol angle')
    def pick_polygons(self, x, y):
        height = self.height()
        with self.lock_glcontext():
            self.pickbuffer.bind()
            index = self.glarea.gl_pick_polygons(x, height-y)
            self.pickbuffer.release()
        if not index:
            self.pickdata = None
            return
        maxis, mslice, mdir, blockpos, symbol, arrowdir = self.glpickdata[index]
        
        if arrowdir is None:
            angle = None
        else:
            angle = self.glarea.get_cursor_angle(x, height-y, arrowdir)
        self.pickdata = self.PickData(maxis, mslice, mdir, blockpos, symbol, angle)
        if DEBUG_DRAW:
            info = {'face': self.model.faces.index(symbol)}
            self.debug_info.emit('pick', info)
        
    def update_selection(self):
        if self.animation_active:
            if self.pickdata is not None:
                self.pickdata = None
                self.set_cursor()
            return
        if self.update_selection_pending:
            return
        self.update_selection_pending = True
        self.timer_update_selection.start()
        
    def on_idle_update_selection(self):
        if self.animation_active:
            if self.pickdata is not None:
                self.pickdata = None
                self.set_cursor()
        else:
            self.pick_polygons(*self.mouse_xy)
            self.set_cursor()
            if DEBUG_DRAW:
                self.update()
        self.update_selection_pending = False
        
    def mouseMoveEvent(self, event):
        self.mouse_xy = event.x(), event.y()
        if DEBUG_MOUSEPOS:
            print(self.mouse_xy)
        
        if not self.button_down_background:
            self.update_selection()
            return
            
        # perform rotation
        offset_x = event.x() - self.last_mouse_x
        offset_y = event.y() - self.last_mouse_y
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(
                                                        self.rotation_x + offset_x,
                                                        self.rotation_y + offset_y)
        self.rotation_x -= offset_x
        self.rotation_y -= offset_y
        self.update()
        
    def mousePressEvent(self, event):
        if self.pickdata is not None:
            if self.animation_active:
                return
            # make a move
            if self.editing_model:
                if event.modifiers() & Qt.ControlModifier:
                    if event.button() == Qt.LeftButton:
                        self.request_rotate_block.emit(self.pickdata.blockpos, False)
                    elif event.button() == Qt.RightButton:
                        self.request_rotate_block.emit(self.pickdata.blockpos, True)
                else:
                    if event.button() == Qt.LeftButton:
                        self.request_swap_blocks.emit(self.pickdata.blockpos,
                                    self.pickdata.maxis, self.pickdata.mslice, self.pickdata.mdir)
            else:
                mslice = -1 if event.modifiers() & Qt.ControlModifier else self.pickdata.mslice
                if event.button() == Qt.LeftButton:
                    self.request_rotation.emit(self.pickdata.maxis, mslice, self.pickdata.mdir)
                elif event.button() == Qt.RightButton and settings.draw.selection_nick == 'simple':
                    self.request_rotation.emit(self.pickdata.maxis, mslice, not self.pickdata.mdir)
        elif event.button() == Qt.LeftButton:
            # start rotation
            self.button_down_background = True
            self.last_mouse_x = event.x()
            self.last_mouse_y = event.y()
        self.update()
        
    def mouseReleaseEvent(self, event):
        if event.button() != Qt.LeftButton:
            return
            
        if self.button_down_background:
            # end rotation
            self.rotation_x += event.x() - self.last_mouse_x
            self.rotation_y += event.y() - self.last_mouse_y
            self.button_down_background = False
            self.update_selection()
            
    def wheelEvent(self, event):
        zoom = settings.draw.zoom * math.pow(1.1, -event.angleDelta().y() / 120)
        zoom_min, zoom_max = settings.draw.zoom_range
        if zoom < zoom_min:
            zoom = zoom_min
        if zoom > zoom_max:
            zoom = zoom_max
        settings.draw.zoom = zoom
            
    def dragEnterEvent(self, event):
        mime_data = event.mimeData()
        debug('drag enter:', mime_data.formats())
        if mime_data.hasColor():
            event.acceptProposedAction()
        elif mime_data.hasUrls() and mime_data.urls()[0].isLocalFile():
            event.acceptProposedAction()
            
    def dropEvent(self, event):
        # when a drag is in progress the pickdata is not updated, so do it now
        self.pick_polygons(event.pos().x(), event.pos().y())
        
        mime_data = event.mimeData()
        if mime_data.hasColor():
            color = mime_data.colorData()
            if self.pickdata is not None:
                self.drop_color.emit(self.pickdata.blockpos, self.pickdata.symbol, color.name())
            else:
                self.drop_color.emit(-1, '', color.name())
        elif mime_data.hasUrls():
            if self.pickdata is None:
                debug('Background image is not supported.')
                return
            uris = mime_data.urls()
            for uri in uris:
                if not uri.isLocalFile():
                    debug('filename "%s" not found or not a local file.' % uri.toString())
                    continue
                filename = uri.toLocalFile()
                if not filename or not os.path.exists(filename):
                    debug('filename "%s" not found.' % filename)
                    continue
                self.drop_file.emit(self.pickdata.blockpos, self.pickdata.symbol, filename)
                break  # ignore other uris
        debug('Skip mime type:', ' '.join(mime_data.formats()))
        
    def set_cursor(self):
        if self.pickdata is None or self.button_down_background:
            index = -1
        elif self.pickdata.angle is None:
            index = -2
        else:
            index = int((self.pickdata.angle+180) / 22.5 + 0.5) % 16
        self.setCursor(self.cursors[index])
        
    def set_std_cursor(self):
        #FIXME: then there still is the wrong cursor
        QTimer.singleShot(0, self.unsetCursor)
        
    def set_editing_mode(self, enable):
        self.editing_model = enable
        self.set_glmodel_selection_mode(0 if enable else settings.draw.selection)
        self.update_selection()
        
    def set_background_color(self, color):
        rgba = QColor()
        rgba.setNamedColor(color)
        self.glarea.set_background_color(rgba.redF(), rgba.greenF(), rgba.blueF())
        
    def reset_rotation(self):
        '''Reset cube rotation'''
        self.rotation_x, self.rotation_y = self.glarea.set_rotation_xy(*self.model.default_rotation)
        self.update()
        
    def reload_shader(self):
        with self.lock_glcontext():
            self.gl_set_shader()
        self.update()
        
    def on_settings_changed(self, key):
        if key == 'draw.speed':
            self.speed = settings.draw.speed
        elif key == 'draw.shader':
            with self.lock_glcontext():
                self.gl_load_shader()
        elif key == 'draw.samples':
            if self.format().samples():
                samples = 2**settings.draw.samples
                if samples == 1:
                    self.glarea.set_antialiasing(False)
                elif samples == self.format().samples():
                    self.glarea.set_antialiasing(True)
        elif key == 'draw.selection':
            self.set_glmodel_selection_mode(settings.draw.selection)
            self.update_selection()
        elif key in ['draw.mirror_faces', 'draw.mirror_distance']:
            self.set_glmodel_full()
        elif key.startswith('theme.faces.'):
            facekey, faceattr = key.split('.')[2:]
            if faceattr == 'color':
                color = QColor()
                color.setNamedColor(settings.theme.faces[facekey].color)
                self.theme.faces[facekey].color = (color.red(), color.green(), color.blue())
                self.set_glmodel_selection_mode(settings.draw.selection)
            elif faceattr == 'image':
                self.load_face_texture(facekey)
                with self.lock_glcontext():
                    self.gl_set_texture_atlas()
                self.set_glmodel_selection_mode(settings.draw.selection)
            elif faceattr == 'mode':
                self.theme.faces[facekey].imagemode = settings.theme.faces[facekey].mode
                self.set_glmodel_selection_mode(settings.draw.selection)
        elif key == 'theme.bgcolor':
            self.set_background_color(settings.theme.bgcolor)
        elif key == 'draw.zoom':
            self.glarea.set_frustum(0, settings.draw.zoom)
        else:
            debug('Unknown settings key changed:', key)
        self.update()
        
        




#!/usr/bin/python3
#-*- coding:utf-8 -*-

#    Copyright © 2009, 2011-2014  B. Clausius <barcc@gmx.de>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os, re

class Py2pyxParseError (Exception): pass


def main(argv):
    if not (1 < len(argv) <= 4):
        print('usage:', os.path.basename(__file__), 'python-file [pyx-filename [pxd-filename]]',
            file=sys.stderr)
        return 1
        
    arg_src = argv[1]
    arg_dst = argv[2] if len(argv) > 2 else None
    arg_pxd = argv[3] if len(argv) > 3 else None
    if not os.path.exists(arg_src):
        print('error:', arg_src, 'does not exist', file=sys.stderr)
        return 1
    create_pyx(arg_src, arg_dst, arg_pxd)
    return 0
    
def create_pyx(src_path, dst_path, pxd_path):
    errors = 0
    openr = lambda path: open(path, 'rt', encoding='utf-8')
    openw = lambda path: open(path, 'wt', encoding='utf-8') if path else sys.stdout
    next_cntlines, next_pyxpat, next_pxdpat = 0, None, None
    
    # key: px command
    # value:
    #      pattern for pyx-file or None (copy line),
    #      pattern for pxd-file or None (skip line),
    #      count next lines or "count" (use px argument),
    #      next lines replace pattern for pyx-file or None,
    #      next lines replace pattern for pxd-file or None
    fdict = {
                None:   (None,            None,     0,       None,              None),
                '+':    (r'\1\5\6 \2\n',  r'\1\5',  0,       None,              None),
                '-':    (r'\1#\2\5\6\n',  None,     1,       r'\1#\2\5\6\n',    r'\1\5'),
                ':':    (r'\1#\2\5\6\n',  None,     '*',     r'\1#\2\5\6\n',    r'\1\5'),
                '.':    (r'\1#\2\5\6\n',  None,     0,       None,              None),
                '/':    (r'\1\5\6 \2\n',  r'\1\5',  1,       r'\1#\2\5\6\n',    None),
                '>':    (r'\1#\2\5\6\n',  r'\1\5',  0,       None,              None),
            }
    fdict['#'] = fdict['>']
    
    # groups in match objects and patterns
    #   1: leading space
    #   2: px command
    #   3: pxd flag
    #   4: px key
    #   5: px argument
    #   6: trailing ":"
    
    pxd_flag = False
    with openr(src_path) as srcf, openw(dst_path) as dstf, openw(pxd_path) as pxdf:
        pxdf.write('#file: {}\n\n'.format(src_path))
        for lineno, line in enumerate(srcf):
            try:
                match = re.match(r'^( *)(#px *(d?)(.)|)(.*?)(:?)\n$', line)
                if match.group(2):
                    if next_cntlines == 0:
                        pxd_flag = match.group(3) == 'd'
                    elif next_cntlines == 1:
                        raise Py2pyxParseError('#px-command found, normal line expected')
                    elif next_cntlines == '*':
                        if match.group(4) != '.':
                            raise Py2pyxParseError('closing #px-command or normal line expected')
                    else:
                        assert False
                    pyxpat, pxdpat, next_cntlines, next_pyxpat, next_pxdpat = fdict[match.group(4)]
                elif next_cntlines == 0:
                    pyxpat, pxdpat, next_cntlines, next_pyxpat, next_pxdpat = fdict[match.group(4)]
                elif next_cntlines == 1:
                    pyxpat = next_pyxpat
                    pxdpat = next_pxdpat
                    next_cntlines = 0
                elif next_cntlines == '*':
                    pyxpat = next_pyxpat
                    pxdpat = next_pxdpat
            except Exception as e:
                print('%s:%d:'%(src_path, lineno+1), e, file=sys.stderr)
                print('  Invalid line:', line, file=sys.stderr, end='')
                pyxpat = None
                pxdpat = None
                next_cntlines = 0
                errors += 1
            finally:
                pyxline = line if pyxpat is None else match.expand(pyxpat)
                pxdline = pxdpat and match.expand(pxdpat)
                dstf.write(pyxline)
                if pxd_flag and pxdline is not None:
                    pxdline = pxdline.rstrip() + '    #line {}\n'.format(lineno+1)
                    pxdf.write(pxdline)
    if errors:
        raise Py2pyxParseError('create_pyx failed with %s errors' % errors)
        
    
if __name__ == '__main__':
    sys.exit(main(sys.argv))
    

